"""Compressors: classes that implement the CompressorInterface."""

import gzip
import zlib

from .interfaces import CompressorInterface


class ZlibCompressor(CompressorInterface):
    """Zlib compressor."""

    @property
    def default_compression_level(self) -> int:
        """Get the default compression level."""
        return zlib.Z_DEFAULT_COMPRESSION

    def compress(self, data: bytes, *, level: int) -> bytes:
        """Compress given data using zlib.

        Args:
            data: Data to compress.

        Keyword Args:
            level: Desired compression level.

        Returns:
            Raw compressed data.
        """
        return zlib.compress(data, level=level)

    def decompress(self, data: bytes) -> bytes:
        """Decompress given compressed data compressed with zlib.

        Args:
            data: Compressed data to decompress.

        Returns:
            Original data.
        """
        return zlib.decompress(data)


class GzipCompressor(CompressorInterface):
    """Gzip compressor."""

    @property
    def default_compression_level(self) -> int:
        """Get the default compression level."""
        return 9  # According to https://docs.python.org/3/library/gzip.html

    def compress(self, data: bytes, *, level: int) -> bytes:
        """Compress given data using gzip.

        Args:
            data: Data to compress.

        Keyword Args:
            level: Desired compression level.

        Returns:
            Raw compressed data.
        """
        return gzip.compress(data, compresslevel=level)

    def decompress(self, data: bytes) -> bytes:
        """Decompress given compressed data compressed with gzip.

        Args:
            data: Compressed data to decompress.

        Returns:
            Original data.
        """
        return gzip.decompress(data)



"""Base classes to test signers."""

import hashlib
import re
import typing
from abc import ABC
from abc import abstractmethod

import pytest
from packaging.version import parse as version_parse

from .. import errors
from ..bases import Blake2Signature
from ..bases import Blake2SignatureDump
from ..bases import HasherChoice
from ..bases import Secret
from ..encoders import B32Encoder
from ..encoders import B64URLEncoder
from ..encoders import HexEncoder
from ..hashers import BLAKE2Hasher
from ..hashers import BLAKE3Hasher
from ..hashers import BLAKEHasher
from ..hashers import has_blake3
from ..interfaces import EncoderInterface
from ..signers import Blake2SerializerSigner
from ..signers import Blake2Signer
from ..signers import Blake2TimestampSigner

Signer = typing.Union[Blake2Signer, Blake2TimestampSigner, Blake2SerializerSigner]

Signature = typing.Union[Blake2SignatureDump, Blake2Signature]


class BaseTests(ABC):
    """Base class for tests."""

    secret = b'0123456789012345'
    person = b'acab'

    @property
    def data(self) -> typing.AnyStr:
        """Get test data."""
        data = b'datadata'
        return data if self.signature_type is bytes else data.decode()

    @property
    @abstractmethod
    def signer_class(self) -> typing.Type[Signer]:
        """Get the signer class to test."""

    def signer(
        self,
        secret: typing.Union[Secret, typing.Sequence[Secret], None] = None,
        **kwargs: typing.Any,
    ) -> Signer:
        """Get the signer to test."""
        return self.signer_class(self.secret if secret is None else secret, **kwargs)

    @property
    @abstractmethod
    def signature_type(self) -> typing.Type[typing.Union[str, bytes]]:
        """Get the signature type for the signer, either `str` or `bytes`."""

    @typing.overload
    def sign(
        self,
        signer: Blake2Signer,
        data: typing.Any,
        **kwargs: typing.Any,
    ) -> bytes:
        """Sign data with the signer."""

    @typing.overload
    def sign(
        self,
        signer: Blake2TimestampSigner,
        data: typing.Any,
        **kwargs: typing.Any,
    ) -> bytes:
        """Sign data with the signer."""

    @typing.overload
    def sign(
        self,
        signer: Blake2SerializerSigner,
        data: typing.Any,
        **kwargs: typing.Any,
    ) -> str:
        """Sign data with the signer."""

    @abstractmethod
    def sign(
        self,
        signer: Signer,
        data: typing.Any,
        **kwargs: typing.Any,
    ) -> typing.AnyStr:
        """Sign data with the signer."""

    @typing.overload
    def unsign(
        self,
        signer: Blake2Signer,
        signed_data: typing.AnyStr,
        **kwargs: typing.Any,
    ) -> bytes:
        """Unsign signed data with the signer."""

    @typing.overload
    def unsign(
        self,
        signer: Blake2TimestampSigner,
        signed_data: typing.AnyStr,
        **kwargs: typing.Any,
    ) -> bytes:
        """Unsign signed data with the signer."""

    @typing.overload
    def unsign(
        self,
        signer: Blake2SerializerSigner,
        signed_data: typing.AnyStr,
        **kwargs: typing.Any,
    ) -> typing.Any:
        """Unsign signed data with the signer."""

    @abstractmethod
    def unsign(
        self,
        signer: Signer,
        signed_data: typing.AnyStr,
        **kwargs: typing.Any,
    ) -> typing.Any:
        """Unsign signed data with the signer."""

    @typing.overload
    def sign_parts(
        self,
        signer: Blake2Signer,
        data: typing.Any,
        **kwargs: typing.Any,
    ) -> Blake2Signature:
        """Sign data with the signer in parts."""

    @typing.overload
    def sign_parts(
        self,
        signer: Blake2TimestampSigner,
        data: typing.Any,
        **kwargs: typing.Any,
    ) -> Blake2Signature:
        """Sign data with the signer in parts."""

    @typing.overload
    def sign_parts(
        self,
        signer: Blake2SerializerSigner,
        data: typing.Any,
        **kwargs: typing.Any,
    ) -> Blake2SignatureDump:
        """Sign data with the signer in parts."""

    @abstractmethod
    def sign_parts(
        self,
        signer: Signer,
        data: typing.Any,
        **kwargs: typing.Any,
    ) -> Signature:
        """Sign data with the signer in parts."""

    @typing.overload
    def unsign_parts(
        self,
        signer: Blake2Signer,
        signature: Signature,
        **kwargs: typing.Any,
    ) -> bytes:
        """Unsign signature with the signer."""

    @typing.overload
    def unsign_parts(
        self,
        signer: Blake2TimestampSigner,
        signature: Signature,
        **kwargs: typing.Any,
    ) -> bytes:
        """Unsign signature with the signer."""

    @typing.overload
    def unsign_parts(
        self,
        signer: Blake2SerializerSigner,
        signature: Signature,
        **kwargs: typing.Any,
    ) -> typing.Any:
        """Unsign signature with the signer."""

    @abstractmethod
    def unsign_parts(
        self,
        signer: Signer,
        signature: Signature,
        **kwargs: typing.Any,
    ) -> typing.Any:
        """Unsign signature with the signer."""

    @staticmethod
    def trick_sign(signer: Signer, data: typing.AnyStr) -> bytes:
        """Sign data properly as if using `Blake2Signer.sign()`.

        This function is useful to trick a signer into unsigning something that
        normally wouldn't be possible because of different safeguard checks.

        Args:
            signer: The signer to use.
            data: data to be signed.

        Returns:
            Signed data, like `signer.sign(data)`, but bypassing checks if any.
        """
        # noinspection PyProtectedMember
        data_b = signer._force_bytes(data)

        # noinspection PyProtectedMember
        return signer._compose(data_b, signature=signer._sign(data_b))

    @pytest.fixture
    def signer_min_digest_size_changed(self) -> typing.Type[Signer]:
        """Fixture to obtain a signer class with a non-default min digest size limit."""
        signer_class = self.signer_class
        min_digest_size = signer_class.MIN_DIGEST_SIZE

        signer_class.MIN_DIGEST_SIZE = 4
        yield signer_class

        signer_class.MIN_DIGEST_SIZE = min_digest_size

    @pytest.fixture
    def signer_min_secret_size_changed(self) -> typing.Type[Signer]:
        """Fixture to obtain a signer class with a non-default min secret size limit."""
        signer_class = self.signer_class
        min_secret_size = signer_class.MIN_SECRET_SIZE

        signer_class.MIN_SECRET_SIZE = 4
        yield signer_class

        signer_class.MIN_SECRET_SIZE = min_secret_size

    def mix_signers_sign_unsign(self, signer1: Signer, signer2: Signer) -> None:
        """Sign and unsign mixing signers to check it fails correctly."""
        signed1 = self.sign(signer1, self.data)
        with pytest.raises(errors.InvalidSignatureError) as exc:
            self.unsign(signer2, signed1)
        assert exc.value.__cause__ is None

        signed2 = self.sign(signer2, self.data)
        with pytest.raises(errors.InvalidSignatureError) as exc:
            self.unsign(signer1, signed2)
        assert exc.value.__cause__ is None

    def test_defaults(self) -> None:
        """Test defaults are correct and work as-is."""
        signer = self.signer()

        assert isinstance(signer, self.signer_class)
        assert isinstance(signer._hasher, BLAKE2Hasher)
        assert signer._hasher._hasher == hashlib.blake2b

        signed = self.sign(signer, self.data)
        assert isinstance(signed, self.signature_type)
        assert self.unsign(signer, signed) == self.data

        signature = self.sign_parts(signer, self.data)
        if self.signature_type is str:
            assert isinstance(signature, Blake2SignatureDump)
        else:
            assert isinstance(signature, Blake2Signature)
        assert self.unsign_parts(signer, signature) == self.data

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        ('hasher', 'hasher_class'),
        (
            (HasherChoice.blake2b, BLAKE2Hasher),
            (HasherChoice.blake2s, BLAKE2Hasher),
            (HasherChoice.blake3, BLAKE3Hasher),
        ),
    )
    def test_hasher_can_be_changed(
        self,
        hasher: HasherChoice,
        hasher_class: typing.Type[BLAKEHasher],
    ) -> None:
        """Test that the hasher can be changed and works correctly."""
        signer = self.signer(hasher=hasher)

        assert isinstance(signer, self.signer_class)
        assert isinstance(signer._hasher, hasher_class)

        signed = self.sign(signer, self.data)
        assert isinstance(signed, self.signature_type)
        assert self.unsign(signer, signed) == self.data

    def test_wrong_hasher_choice(self) -> None:
        """Test wrong hasher choice."""
        with pytest.raises(errors.InvalidOptionError, match='invalid hasher choice'):
            self.signer(hasher='blake2')

    def test_all_instantiation_params_can_be_changed(self) -> None:
        """Test correct signing and unsigning using all non-default options."""
        signer = self.signer(
            self.secret,
            personalisation=self.person,
            hasher=self.signer_class.Hashers.blake2s,
            digest_size=32,
            deterministic=True,
            separator=',',
            encoder=B32Encoder,
        )

        assert isinstance(signer, self.signer_class)

        signed = self.sign(signer, self.data)
        assert isinstance(signed, self.signature_type)
        assert self.unsign(signer, signed) == self.data

        signature = self.sign_parts(signer, self.data)
        assert isinstance(signature.data, self.signature_type)
        assert isinstance(signature.signature, self.signature_type)
        if self.signature_type is str:
            assert isinstance(signature, Blake2SignatureDump)
        else:
            assert isinstance(signature, Blake2Signature)
            assert self.data == signature.data

        assert self.unsign_parts(signer, signature) == self.data

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_sign_unsign_with_person(self, hasher: HasherChoice) -> None:
        """Test signing and unsigning using person is correct."""
        signer = self.signer(personalisation=self.person, hasher=hasher)

        assert self.data == self.unsign(signer, self.sign(signer, self.data))

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_sign_unsign_with_different_person(self, hasher: HasherChoice) -> None:
        """Test signing and unsigning using different person fails correctly."""
        signer1 = self.signer(personalisation=self.person, hasher=hasher)
        signer2 = self.signer(personalisation=self.person * 2, hasher=hasher)

        self.mix_signers_sign_unsign(signer1, signer2)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_sign_unsign_with_different_encoder(self, hasher: HasherChoice) -> None:
        """Test signing and unsigning using different encoder fails correctly."""
        signer1 = self.signer(encoder=B64URLEncoder, hasher=hasher)
        signer2 = self.signer(encoder=B32Encoder, hasher=hasher)

        self.mix_signers_sign_unsign(signer1, signer2)

    @pytest.mark.parametrize(
        ('secret', 'person', 'separator', 'data'),
        (
            ('setec astronomy.', 'acab', ',', 'datadata'),
            (b'setec astronomy.', 'acab', ',', 'datadata'),
            ('setec astronomy.', b'acab', ',', 'datadata'),
            ('setec astronomy.', 'acab', b',', 'datadata'),
            ('setec astronomy.', b'acab', b',', 'datadata'),
            (b'setec astronomy.', b'acab', b',', 'datadata'),
        ),
    )
    def test_string_instead_of_bytes_inputs(
        self,
        secret: typing.Union[str, bytes],
        person: typing.Union[str, bytes],
        separator: typing.Union[str, bytes],
        data: typing.Union[str, bytes],
    ) -> None:
        """Test non-bytes values for parameters such as secret, person, data, etc."""
        signer = self.signer(
            secret,
            personalisation=person,
            separator=separator,
        )
        assert isinstance(signer, self.signer_class)

        signed = self.sign(signer, data)
        assert isinstance(signed, self.signature_type)
        if isinstance(separator, self.signature_type):
            assert separator in signed
        elif isinstance(separator, bytes):
            assert separator.decode() in signed
        else:
            assert separator.encode() in signed

        unsigned = self.unsign(signer, signed)
        if isinstance(data, self.signature_type):
            assert data == unsigned
        else:
            assert data == unsigned.decode()

        # Now change `signed` type and check if it works correctly
        if self.signature_type is bytes:
            unsigned = self.unsign(signer, signed.decode())
        else:
            unsigned = self.unsign(signer, signed.encode())
        if isinstance(data, self.signature_type):
            assert data == unsigned
        else:
            assert data == unsigned.decode()

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_sign_is_unique_non_deterministic(self, hasher: HasherChoice) -> None:
        """Test that each signing is unique because of salt."""
        signer = self.signer(deterministic=False, hasher=hasher)

        signed1 = self.sign(signer, self.data)
        signed2 = self.sign(signer, self.data)

        assert len(signed1) == len(signed2)
        assert type(signed1) == type(signed2)
        assert signed1[-len(self.data):] == signed2[-len(self.data):]
        assert signed1 != signed2

        unsigned1 = self.unsign(signer, signed1)
        assert self.data == unsigned1

        unsigned2 = self.unsign(signer, signed2)
        assert self.data == unsigned2

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_sign_unsign_deterministic(self, hasher: HasherChoice) -> None:
        """Test sign and unsign with a deterministic signature."""
        signer = self.signer(deterministic=True, hasher=hasher)

        signed1 = self.sign(signer, self.data)
        signed2 = self.sign(signer, self.data)
        assert signed1 == signed2

        unsigned = self.unsign(signer, signed1)
        assert self.data == unsigned

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_sign_parts_is_unique_non_deterministic(self, hasher: HasherChoice) -> None:
        """Test that each signing in parts is unique because of salt."""
        signer = self.signer(deterministic=False, hasher=hasher)

        signature1 = self.sign_parts(signer, self.data)
        signature2 = self.sign_parts(signer, self.data)
        assert signature1 != signature2

        unsigned = self.unsign_parts(signer, signature1)
        assert self.data == unsigned

        unsigned = self.unsign_parts(signer, signature2)
        assert self.data == unsigned

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_sign_unsign_parts_deterministic(self, hasher: HasherChoice) -> None:
        """Test signing and unsigning in parts deterministically."""
        signer = self.signer(deterministic=True, hasher=hasher)

        signature1 = self.sign_parts(signer, self.data)
        signature2 = self.sign_parts(signer, self.data)
        assert signature1 == signature2

        unsigned = self.unsign_parts(signer, signature1)
        assert self.data == unsigned

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    @pytest.mark.parametrize('separator', (':', b':'))
    def test_separator_can_be_changed(
        self,
        separator: typing.Union[str, bytes],
        hasher: HasherChoice,
    ) -> None:
        """Test that the separator can be changed."""
        signer = self.signer(separator=separator, hasher=hasher)

        signed = self.sign(signer, self.data)
        if isinstance(separator, self.signature_type):
            assert separator in signed
        elif isinstance(separator, bytes):
            assert separator.decode() in signed
        else:
            assert separator.encode() in signed

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    @pytest.mark.parametrize(
        ('encoder', 'regex'),
        (
            (B64URLEncoder, r'^[a-zA-Z0-9_\-]+(\.[a-zA-Z0-9_\-]+)?\.datadata$'),
            (B32Encoder, r'^[A-Z2-7]+(\.[A-Z2-7]+)?\.datadata$'),
            (HexEncoder, r'^[A-F0-9]+(\.[A-F0-9]+)?\.datadata$'),
        ),
    )
    def test_sign_unsign_with_encoder(
        self,
        encoder: typing.Type[EncoderInterface],
        regex: str,
        hasher: HasherChoice,
    ) -> None:
        """Test signing and unsigning using an encoder."""
        data = 'datadata' if self.signature_type is str else b'datadata'
        signer = self.signer(encoder=encoder, hasher=hasher)

        signed = self.sign(signer, data)
        assert re.match(
            regex,
            signed if self.signature_type is str else signed.decode(),
        ), signed

        unsigned = self.unsign(signer, signed)
        assert data == unsigned

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_sign_unsign_parts_both_sig_containers(self, hasher: HasherChoice) -> None:
        """Test signing and unsigning in parts accepting both signature containers."""
        signer = self.signer(hasher=hasher)

        signature = self.sign_parts(signer, self.data)
        if self.signature_type is str:
            assert isinstance(signature, Blake2SignatureDump)
        else:
            assert isinstance(signature, Blake2Signature)

        if self.signature_type is str:
            other_signature = Blake2Signature(
                data=signature.data.encode(),
                signature=signature.signature.encode(),
            )
        else:
            other_signature = Blake2SignatureDump(
                data=signature.data.decode(),
                signature=signature.signature.decode(),
            )
        unsigned = self.unsign_parts(signer, other_signature)
        assert self.data == unsigned

    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_secret_too_short(self, hasher: HasherChoice) -> None:
        """Test secret too short."""
        with pytest.raises(
                errors.InvalidOptionError,
                match='1st secret should be longer than',
        ):
            self.signer(b's' * (self.signer_class.MIN_SECRET_SIZE - 1), hasher=hasher)

    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_secret_too_short_in_sequence(self, hasher: HasherChoice) -> None:
        """Test secret too short in a sequence."""
        with pytest.raises(
                errors.InvalidOptionError,
                match='1st secret should be longer than',
        ):
            self.signer([b's' * (self.signer_class.MIN_SECRET_SIZE - 1)], hasher=hasher)

        with pytest.raises(
                errors.InvalidOptionError,
                match='2nd secret should be longer than',
        ):
            self.signer(
                [self.secret, b's' * (self.signer_class.MIN_SECRET_SIZE - 1)],
                hasher=hasher,
            )

    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_digest_too_small(self, hasher: HasherChoice) -> None:
        """Test digest too small."""
        sizes = (
            -1,
            self.signer_class.MIN_DIGEST_SIZE - 1,
        )
        for digest_size in sizes:
            with pytest.raises(
                    errors.InvalidOptionError,
                    match='digest_size should be',
            ):
                self.signer(hasher=hasher, digest_size=digest_size)

    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            # blake3 has no digest limit
        ),
    )
    def test_digest_too_large(self, hasher: HasherChoice) -> None:
        """Test digest too large."""
        digest_size = getattr(hashlib, hasher.value).MAX_DIGEST_SIZE + 1
        with pytest.raises(
                errors.InvalidOptionError,
                match='digest_size should be',
        ):
            self.signer(hasher=hasher, digest_size=digest_size)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    def test_blake3_no_digest_size_limit(self) -> None:
        """Test that BLAKE3 has no digest size limit."""
        signer = self.signer(
            self.secret,
            digest_size=1_000,
            hasher=self.signer_class.Hashers.blake3,
        )

        signed = self.sign(signer, self.data)
        assert len(signed) > 1_000
        assert self.data == self.unsign(signer, signed)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_unsign_no_separator(self, hasher: HasherChoice) -> None:
        """Test unsign with wrong data without separator."""
        signer = self.signer(hasher=hasher)

        with pytest.raises(errors.SignatureError, match='separator not found') as exc:
            self.unsign(signer, b'12345678')
        assert exc.value.__cause__ is None

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_unsign_short_data_without_signature(self, hasher: HasherChoice) -> None:
        """Test unsign with very short data without signature."""
        signer = self.signer(hasher=hasher)

        with pytest.raises(
                errors.SignatureError,
                match='signature information is missing',
        ) as exc:
            self.unsign(
                signer,
                b'.',  # The shortest possible that passes the separator check
            )
        assert exc.value.__cause__ is None

        with pytest.raises(
                errors.SignatureError,
                match='signature information is missing',
        ) as exc:
            self.unsign(signer, b'.12345678')
        assert exc.value.__cause__ is None

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_unsign_wrong_signature(self, hasher: HasherChoice) -> None:
        """Test unsign with wrong signed data."""
        signer = self.signer(hasher=hasher)

        with pytest.raises(
                errors.InvalidSignatureError,
                match='signature is not valid',
        ) as exc:
            self.unsign(signer, b's.')
        assert exc.value.__cause__ is None

        signed = self.sign(signer, self.data)

        with pytest.raises(
                errors.InvalidSignatureError,
                match='signature is not valid',
        ) as exc:
            self.unsign(signer, signed[1:])
        assert exc.value.__cause__ is None

        with pytest.raises(
                errors.InvalidSignatureError,
                match='signature is not valid',
        ) as exc:
            self.unsign(signer, signed[:-1])
        assert exc.value.__cause__ is None

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    @pytest.mark.parametrize(
        ('encoder', 'separator'),
        (
            (B64URLEncoder, b'A'),
            (B32Encoder, b'A'),
            (HexEncoder, b'A'),
        ),
    )
    def test_separator_in_encoder_alphabet(
        self,
        encoder: typing.Type[EncoderInterface],
        separator: bytes,
        hasher: HasherChoice,
    ) -> None:
        """Test error occurs when the separator is in the encoder alphabet."""
        with pytest.raises(
                errors.InvalidOptionError,
                match='separator character must not belong to the encoder',
        ):
            self.signer(separator=separator, encoder=encoder, hasher=hasher)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_separator_non_ascii(self, hasher: HasherChoice) -> None:
        """Test error occurs when the separator is non-ascii."""
        with pytest.raises(
                errors.InvalidOptionError,
                match='separator character must be ASCII',
        ):
            self.signer(separator=b'\x87', hasher=hasher)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_separator_empty(self, hasher: HasherChoice) -> None:
        """Test error occurs when the separator is empty."""
        with pytest.raises(
                errors.InvalidOptionError,
                match='the separator character must have a value',
        ):
            self.signer(separator=b'', hasher=hasher)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_custom_encoder_non_ascii_alphabet(self, hasher: HasherChoice) -> None:
        """Test encoder having a non-ASCII alphabet raises exception."""

        class Encoder(EncoderInterface):
            """Wrong encoder."""

            @property
            def alphabet(self) -> bytes:
                """Get encoder alphabet."""
                return b'\x87'

            def encode(self, data: typing.AnyStr) -> bytes:
                """Encode data."""
                pass  # pragma: nocover

            def decode(self, data: typing.AnyStr) -> bytes:
                """Decode data."""
                pass  # pragma: nocover

        with pytest.raises(
                errors.InvalidOptionError,
                match='encoder alphabet must be ASCII',
        ):
            self.signer(encoder=Encoder, hasher=hasher)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_custom_encoder_empty_alphabet(self, hasher: HasherChoice) -> None:
        """Test encoder having an empty alphabet raises exception."""

        class Encoder(EncoderInterface):
            """Wrong encoder."""

            @property
            def alphabet(self) -> bytes:
                """Get encoder alphabet."""
                return b''

            def encode(self, data: typing.AnyStr) -> bytes:
                """Encode data."""
                pass  # pragma: nocover

            def decode(self, data: typing.AnyStr) -> bytes:
                """Decode data."""
                pass  # pragma: nocover

        with pytest.raises(
                errors.InvalidOptionError,
                match='encoder alphabet must have a value',
        ):
            self.signer(encoder=Encoder, hasher=hasher)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_minimum_digest_size_can_be_changed(
        self,
        signer_min_digest_size_changed,
        hasher: HasherChoice,
    ) -> None:
        """Test that the min digest size limit can be changed."""
        signer = signer_min_digest_size_changed(
            self.secret,
            digest_size=8,
            hasher=hasher,
        )

        assert self.data == self.unsign(signer, self.sign(signer, self.data))

        with pytest.raises(
                errors.InvalidOptionError,
                match='digest_size should be',
        ):
            signer_min_digest_size_changed(
                self.secret,
                digest_size=signer_min_digest_size_changed.MIN_DIGEST_SIZE - 1,
                hasher=hasher,
            )

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_minimum_secret_size_can_be_changed(
        self,
        signer_min_secret_size_changed,
        hasher: HasherChoice,
    ) -> None:
        """Test that the min secret size limit can be changed."""
        signer = signer_min_secret_size_changed(b'12345678', hasher=hasher)

        assert self.data == self.unsign(signer, self.sign(signer, self.data))

        with pytest.raises(
                errors.InvalidOptionError,
                match='secret should be longer than',
        ):
            signer_min_secret_size_changed(
                b's' * (signer_min_secret_size_changed.MIN_SECRET_SIZE - 1),
                hasher=hasher,
            )

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    @pytest.mark.parametrize(
        'secret',
        (
            'a string secret' * 3,
            b'a bytes secret' * 3,
            ['a list' * 3, 'string secret' * 3],
            [b'a list' * 3, b'bytes secret' * 3],
            ['a list' * 3, b'mixed secret' * 3],
            ('a tuple' * 3, 'secret' * 3),
        ),
    )
    def test_secret_can_be_any_sequence(
        self,
        secret: typing.Union[Secret, typing.Sequence[Secret]],
        hasher: HasherChoice,
    ) -> None:
        """Test that the min digest size limit can be changed."""
        signer = self.signer(secret, hasher=hasher)
        assert self.data == self.unsign(signer, self.sign(signer, self.data))

    @abstractmethod
    def test_versions_compat(
        self,
        version: str,
        hasher: HasherChoice,
        signed: str,
        compat: bool,
    ) -> None:
        """Test if previous versions' signed data is compatible with the current one."""
        signer: Signer = self.signer(b'too many secrets!', hasher=hasher)
        data: bytes = b'is compat ensured?'

        assert version_parse(version) >= version_parse('1.2.1')

        if compat:
            unsigned = self.unsign(signer, signed)

            if self.signature_type is bytes:
                assert data == unsigned
            else:
                assert data.decode() == unsigned
        else:
            with pytest.raises(errors.SignedDataError):
                self.unsign(signer, signed)



"""Serializer signers tests."""

import io
import json
import typing
import zlib
from abc import ABC
from datetime import timedelta
from secrets import token_bytes
from time import time
from unittest import mock

import pytest

from .bases import BaseTests
from .bases import Signature
from .bases import Signer
from .test_timestampsigner import TimestampSignerTestsBase
from .. import errors
from ..compressors import GzipCompressor
from ..compressors import ZlibCompressor
from ..encoders import B32Encoder
from ..encoders import B64URLEncoder
from ..encoders import HexEncoder
from ..hashers import HasherChoice
from ..hashers import has_blake3
from ..interfaces import CompressorInterface
from ..interfaces import EncoderInterface
from ..serializers import JSONSerializer
from ..serializers import NullSerializer
from ..signers import Blake2SerializerSigner


class SerializerSignerTestsBase(BaseTests, ABC):
    """Base to test a serializer signer."""

    data_compressible = 'datadata' * 100  # so compression is meaningful
    data_incompressible = token_bytes(10).hex()  # so it can't be compressed

    @property
    def signer_class(self) -> typing.Type[Signer]:
        """Get the signer class to test."""
        return Blake2SerializerSigner

    @property
    def signature_type(self) -> typing.Type[typing.Union[str, bytes]]:
        """Get the signature type for the signer (`str`)."""
        return str

    def sign(
        self,
        signer: Signer,
        data: typing.Any,
        **kwargs: typing.Any,
    ) -> typing.AnyStr:
        """Sign data with the signer."""
        return signer.dumps(data, **kwargs)

    def unsign(
        self,
        signer: Signer,
        signed_data: typing.Union[str, bytes],
        **kwargs: typing.Any,
    ) -> typing.Any:
        """Unsign signed data with the signer."""
        return signer.loads(signed_data)

    def sign_parts(
        self,
        signer: Signer,
        data: typing.Any,
        **kwargs: typing.Any,
    ) -> Signature:
        """Sign data with the signer in parts."""
        return signer.dumps_parts(data, **kwargs)

    def unsign_parts(
        self,
        signer: Signer,
        signature: Signature,
        **kwargs: typing.Any,
    ) -> typing.Any:
        """Unsign signature with the signer."""
        return signer.loads_parts(signature)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    @pytest.mark.parametrize(
        'data',
        (
            None,
            True,
            'datadata',
            [1, 2, 3],
            {
                'user': 'hackan',
            },
            [],
            {},
        ),
    )
    def test_dumps_loads_default(
        self,
        data: typing.Any,
        hasher: HasherChoice,
    ) -> None:
        """Test dumping and loading with defaults is correct."""
        signer = self.signer(hasher=hasher)

        signed = self.sign(signer, data)
        assert isinstance(signed, str)

        unsigned = self.unsign(signer, signed)
        assert data == unsigned

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    @pytest.mark.parametrize(
        ('encoder', 'regex'),
        (
            (B64URLEncoder, r'^[a-zA-Z0-9_\-]+(\.[a-zA-Z0-9_\-]+)?\.[a-zA-Z0-9_\-]+$'),
            (B32Encoder, r'^[A-Z2-7]+(\.[A-Z2-7]+)?\.[A-Z2-7]+$'),
            (HexEncoder, r'^[A-F0-9]+(\.[A-F0-9]+)?\.[A-F0-9]+$'),
        ),
    )
    def test_sign_unsign_with_encoder(
        self,
        encoder: typing.Type[EncoderInterface],
        regex: str,
        hasher: HasherChoice,
    ) -> None:
        """Test signing and unsigning using an encoder."""
        super().test_sign_unsign_with_encoder(encoder, regex, hasher)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_dumps_loads_compression_level_default(self, hasher: HasherChoice) -> None:
        """Test dumping and loading with default compression level is correct."""
        signer = self.signer(hasher=hasher)

        signed = self.sign(signer, self.data_compressible, compress=False)
        signed_compressed = self.sign(
            signer,
            self.data_compressible,
            compress=True,
            compression_level=None,
        )
        assert len(signed_compressed) < len(signed)

        assert self.data_compressible == self.unsign(signer, signed_compressed)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_dumps_loads_compression_level(self, hasher: HasherChoice) -> None:
        """Test dumping and loading changing compression level is correct."""
        signer = self.signer(hasher=hasher)

        signed = self.sign(signer, self.data_compressible, compress=False)
        signed_compressed1 = self.sign(
            signer,
            self.data_compressible,
            compress=True,
            compression_level=1,
        )
        signed_compressed2 = self.sign(
            signer,
            self.data_compressible,
            compress=True,
            compression_level=9,
        )
        assert len(signed_compressed2) < len(signed_compressed1) < len(signed)

        assert self.data_compressible == self.unsign(signer, signed_compressed1)
        assert self.data_compressible == self.unsign(signer, signed_compressed2)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_dumps_loads_auto_compression_incompressible(
        self,
        hasher: HasherChoice,
    ) -> None:
        """Test dumping and loading with auto compression is correct."""
        signer = self.signer(hasher=hasher)

        signed = self.sign(signer, self.data_incompressible, compress=False)
        signed_not_compressed = self.sign(
            signer,
            self.data_incompressible,
            compress=True,
        )
        assert len(signed_not_compressed) == len(signed)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_dumps_loads_force_compression(self, hasher: HasherChoice) -> None:
        """Test dumping and loading forcing compression is correct."""
        signer = self.signer(hasher=hasher)

        signed = self.sign(signer, self.data_compressible, compress=False)
        signed_compressed = self.sign(
            signer,
            self.data_compressible,
            force_compression=True,
        )
        assert len(signed_compressed) < len(signed)

        assert self.data_compressible == self.unsign(signer, signed_compressed)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_dumps_force_compression_bypasses_compress(
        self,
        hasher: HasherChoice,
    ) -> None:
        """Test dumping and loading forcing compression bypasses compress."""
        signer = self.signer(hasher=hasher)

        signed = self.sign(signer, self.data_compressible, compress=False)
        signed_compressed = self.sign(
            signer,
            self.data_compressible,
            compress=False,
            force_compression=True,
        )
        assert len(signed_compressed) < len(signed)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_dumps_force_compression_bypasses_compress_detrimental(
        self,
        hasher: HasherChoice,
    ) -> None:
        """Test dumps forcing compression bypasses compress and is detrimental."""
        signer = self.signer(hasher=hasher)

        signed = self.sign(signer, self.data_incompressible, compress=False)
        signed_compressed = self.sign(
            signer,
            self.data_incompressible,
            compress=False,
            force_compression=True,
        )
        assert len(signed_compressed) > len(signed)

    def test_sign_unsign_with_different_serializer(self) -> None:
        """Test signing and unsigning using different serializer fails correctly."""
        signer1 = self.signer(serializer=JSONSerializer)
        signer2 = self.signer(serializer=NullSerializer)

        self.mix_signers_sign_unsign(signer1, signer2)

    def test_sign_unsign_with_different_compressor(self) -> None:
        """Test signing and unsigning using different compressor fails correctly."""
        signer1 = self.signer(compressor=ZlibCompressor)
        signer2 = self.signer(compressor=GzipCompressor)

        self.mix_signers_sign_unsign(signer1, signer2)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_dumps_loads_with_custom_serializer(  # noqa: C901
            self,
            hasher: HasherChoice,
    ) -> None:
        """Test dumping and loading using a custom serializer."""

        class Justice:
            """Some object."""

            def __init__(self, a: str = ''):
                self.a = a

        class CustomJSONEncoder(json.JSONEncoder):
            """Custom JSON encoder."""

            def default(self, o):
                """Encode object."""
                if isinstance(o, Justice):
                    return o.a

                return super().default(o)  # pragma: no cover

        class MyJSONSerializer(JSONSerializer):
            """Custom JSON serializer."""

            def serialize(self, data: typing.Any, **kwargs: typing.Any) -> bytes:
                """Serialize self.data."""
                return super().serialize(data, cls=CustomJSONEncoder, **kwargs)

        obj = Justice('acab')
        signer = self.signer(serializer=MyJSONSerializer, hasher=hasher)

        unsigned = self.unsign(signer, self.sign(signer, obj))
        assert obj.a == unsigned

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    @pytest.mark.parametrize('flag', ('!', b'!'))
    def test_compression_flag_can_be_changed(
        self,
        flag: typing.Union[str, bytes],
        hasher: HasherChoice,
    ) -> None:
        """Test that the compression flag can be changed."""
        signer = self.signer(compression_flag=flag, hasher=hasher)

        if isinstance(flag, bytes):
            assert signer._compression_flag == flag
        else:
            assert signer._compression_flag == flag.encode()

        signed = self.sign(signer, self.data_compressible)
        if isinstance(flag, bytes):
            assert flag.decode() in signed
        else:
            assert flag in signed

        unsigned = signer._proper_unsign(signer._decompose(signed.encode()))
        if isinstance(flag, bytes):
            assert unsigned.startswith(flag)
        else:
            assert unsigned.startswith(flag.encode())

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_compression_can_not_be_bombed(self, hasher: HasherChoice) -> None:
        """Test that a malicious input can't bomb the compression."""
        bomb_data = b'\x00' * 1048576
        bomb = zlib.compress(bomb_data)

        assert len(bomb) < len(bomb_data)

        signer = self.signer(
            serializer=NullSerializer,
            compressor=ZlibCompressor,
            hasher=hasher,
        )

        payload = signer._compression_flag + bomb
        signed = self.sign(signer, payload, compress=False)
        # If the bomb worked, then when unsigning the payload would be decompressed
        # If it failed, then the payload would stay the same, which is what should
        # happen here
        unsigned = self.unsign(signer, signed)
        assert len(payload) == len(unsigned)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_compression_ratio_can_be_changed(self, hasher: HasherChoice) -> None:
        """Test that the compression ratio can be changed."""
        data = 'datadatadatadata'  # Only somewhat compressible
        signer1 = self.signer(compression_ratio=10, hasher=hasher)
        signer2 = self.signer(compression_ratio=20, hasher=hasher)

        signed1 = self.sign(signer1, data, compress=True)  # Compressed
        signed2 = self.sign(signer2, data, compress=True)  # Not compressed

        assert len(signed1) < len(signed2)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    @pytest.mark.parametrize(
        'compressor',
        (
            ZlibCompressor,
            GzipCompressor,
        ),
    )
    def test_dumps_loads_with_compressor(
        self,
        compressor: typing.Type[CompressorInterface],
        hasher: HasherChoice,
    ) -> None:
        """Test dumping and loading using a compressor."""
        signer = self.signer(compressor=compressor, hasher=hasher)

        signed = self.sign(signer, self.data_compressible, compress=False)
        signed_compressed = self.sign(signer, self.data_compressible, compress=True)
        assert len(signed_compressed) < len(signed)

        unsigned = self.unsign(signer, signed_compressed)
        assert self.data_compressible == unsigned

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    @pytest.mark.parametrize(
        'data',
        ('datadata', b'datadata'),
    )
    def test_dumps_loads_with_null_serializer(
        self,
        data: typing.AnyStr,
        hasher: HasherChoice,
    ) -> None:
        """Test dumping and loading using the null serializer."""
        signer = self.signer(serializer=NullSerializer, hasher=hasher)

        signed = self.sign(signer, data)
        assert isinstance(signed, str)

        unsigned = self.unsign(signer, signed)
        assert isinstance(unsigned, bytes)
        if isinstance(data, bytes):
            assert data == unsigned
        else:
            assert data == unsigned.decode()

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    @pytest.mark.parametrize(
        'file_class',
        (
            io.StringIO,
            io.BytesIO,
        ),
    )
    def test_dump_load_file(
        self,
        file_class: typing.Type[typing.IO],
        hasher: HasherChoice,
    ) -> None:
        """Test dumping and loading to/from a file."""
        file = file_class()
        signer = self.signer(hasher=hasher)

        signed = signer.dump(self.data, file)
        assert file.tell() == len(signed)

        assert self.data == self.unsign(signer, signed)

        file.seek(0)
        unsigned = signer.load(file)
        assert file.tell() == len(signed)
        assert self.data == unsigned

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    @pytest.mark.parametrize(
        'file_class',
        (
            io.StringIO,
            io.BytesIO,
        ),
    )
    def test_dump_load_file_with_null_serializer(
        self,
        file_class: typing.Type[typing.IO],
        hasher: HasherChoice,
    ) -> None:
        """Test dumping and loading to/from a file."""
        file = file_class()
        signer = self.signer(serializer=NullSerializer, hasher=hasher)
        data = self.data.encode()

        signed = signer.dump(data, file)
        assert file.tell() == len(signed)

        assert data == self.unsign(signer, signed)

        file.seek(0)
        unsigned = signer.load(file)
        assert file.tell() == len(signed)
        assert data == unsigned

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    @pytest.mark.parametrize(
        ('file_class', 'initial_data'),
        (
            (io.StringIO, 'acab' * 5),
            (io.BytesIO, b'acab' * 5),
        ),
    )
    def test_dump_load_file_containing_data(
        self,
        file_class: typing.Type[typing.IO],
        initial_data: typing.Union[str, bytes],
        hasher: HasherChoice,
    ) -> None:
        """Test dumping and loading to/from a file that contains data.

        This is important to verify that we are not changing the file cursor in
        any unpredictable way.
        """
        file = file_class()
        file.write(initial_data)
        initial_pos = file.tell()
        signer = self.signer(hasher=hasher)

        signed = signer.dump(self.data, file)
        assert file.tell() == (len(initial_data) + len(signed))

        file.seek(initial_pos)
        unsigned = signer.load(file)
        assert file.tell() == (len(initial_data) + len(signed))
        assert self.data == unsigned

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_dumps_serializer_kwargs(self, hasher: HasherChoice) -> None:
        """Test dumping using serializer kwargs."""
        signer = self.signer(
            deterministic=True,
            serializer=JSONSerializer,
            hasher=hasher,
        )
        data = {
            'a': 'b',
            1: 2,
        }

        signed1 = self.sign(
            signer,
            data,
            serializer_kwargs={'separators': ('.', ';')},
        )
        signed1_1 = self.sign(
            signer,
            data,
            serializer_kwargs={'separators': ('.', ';')},
        )
        assert signed1 == signed1_1  # It is effectively deterministic

        signed2 = self.sign(
            signer,
            data,
            serializer_kwargs={'separators': ('.', ',')},
        )
        assert signed1 != signed2  # Change due only to the serializer options

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_dump_serializer_kwargs(self, hasher: HasherChoice) -> None:
        """Test dumping to file using serializer kwargs."""
        file1 = io.StringIO()
        file1_1 = io.StringIO()
        file2 = io.StringIO()
        signer = self.signer(
            deterministic=True,
            serializer=JSONSerializer,
            hasher=hasher,
        )
        data = {
            'a': 'b',
            1: 2,
        }

        signer.dump(
            data,
            file1,
            serializer_kwargs={'separators': ('.', ';')},
        )
        signer.dump(
            data,
            file1_1,
            serializer_kwargs={'separators': ('.', ';')},
        )
        file1.seek(0)
        file1_1.seek(0)
        assert file1.read() == file1_1.read()  # It is effectively deterministic

        signer.dump(
            data,
            file2,
            serializer_kwargs={'separators': ('.', ',')},
        )
        file1.seek(0)
        file2.seek(0)
        assert file1.read() != file2.read()  # Change due only to the serializer options

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_dumps_parts_serializer_kwargs(self, hasher: HasherChoice) -> None:
        """Test dumping parts using serializer kwargs."""
        signer = self.signer(
            deterministic=True,
            serializer=JSONSerializer,
            hasher=hasher,
        )
        data = {
            'a': 'b',
            1: 2,
        }

        signed1 = self.sign_parts(
            signer,
            data,
            serializer_kwargs={'separators': ('.', ';')},
        )
        signed1_1 = self.sign_parts(
            signer,
            data,
            serializer_kwargs={'separators': ('.', ';')},
        )
        assert signed1 == signed1_1  # It is effectively deterministic

        signed2 = self.sign_parts(
            signer,
            data,
            serializer_kwargs={'separators': ('.', ',')},
        )
        assert signed1 != signed2  # Change due only to the serializer options

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_dumps_unserializable_data(self, hasher: HasherChoice) -> None:
        """Test dumps unserializable data fails correctly."""
        signer = self.signer(serializer=JSONSerializer, hasher=hasher)

        with pytest.raises(
                errors.SerializationError,
                match='data can not be serialized',
        ) as exc:
            self.sign(signer, self.data.encode())  # Any non JSON encodable type
        assert exc.value.__cause__ is not None

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_loads_wrong_data(self, hasher: HasherChoice) -> None:
        """Test loads wrong data."""
        signer = self.signer(hasher=hasher)

        with pytest.raises(
                errors.ConversionError,
                match='value can not be converted to bytes',
        ) as exc:
            self.unsign(signer, 1.0)  # type: ignore
        assert exc.value.__cause__ is not None

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    @mock.patch('blake2signer.utils.base64.urlsafe_b64decode')
    def test_loads_decode_error(
        self,
        mock_b64decode: mock.MagicMock,
        hasher: HasherChoice,
    ) -> None:
        """Test loads wrong data causing decoding error."""
        mock_b64decode.side_effect = ValueError
        signer = self.signer(encoder=B64URLEncoder, hasher=hasher)

        signed = self.sign(signer, self.data)
        with pytest.raises(
                errors.DecodeError,
                match='data can not be decoded',
        ) as exc:
            self.unsign(signer, signed)
        assert isinstance(exc.value.__cause__, ValueError)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    @mock.patch('blake2signer.compressors.zlib.decompress')
    def test_loads_decompression_error(
        self,
        mock_decompress: mock.MagicMock,
        hasher: HasherChoice,
    ) -> None:
        """Test loads wrong data causing decompression error."""
        mock_decompress.side_effect = zlib.error
        signer = self.signer(compressor=ZlibCompressor, hasher=hasher)

        signed = self.sign(signer, self.data_compressible, compress=True)
        with pytest.raises(
                errors.DecompressionError,
                match='data can not be decompressed',
        ) as exc:
            self.unsign(signer, signed)
        assert isinstance(exc.value.__cause__, zlib.error)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    @mock.patch('blake2signer.serializers.json.loads')
    def test_loads_unserialization_error(
        self,
        mock_json_loads: mock.MagicMock,
        hasher: HasherChoice,
    ) -> None:
        """Test loads wrong data causing unserialization error."""
        mock_json_loads.side_effect = ValueError
        signer = self.signer(serializer=JSONSerializer, hasher=hasher)

        signed = self.sign(signer, self.data)
        with pytest.raises(
                errors.UnserializationError,
                match='data can not be unserialized',
        ) as exc:
            self.unsign(signer, signed)
        assert isinstance(exc.value.__cause__, ValueError)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    @mock.patch('blake2signer.compressors.zlib.compress')
    def test_dumps_compression_error(
        self,
        mock_zlib_compress: mock.MagicMock,
        hasher: HasherChoice,
    ) -> None:
        """Test compression error while dumping."""
        mock_zlib_compress.side_effect = zlib.error

        signer = self.signer(compressor=ZlibCompressor, hasher=hasher)

        with pytest.raises(
                errors.CompressionError,
                match='data can not be compressed',
        ) as exc:
            self.sign(signer, self.data_compressible, compress=True)
        assert isinstance(exc.value.__cause__, zlib.error)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    @mock.patch('blake2signer.encoders.b64encode')
    def test_dumps_encoding_error(
        self,
        mock_b64encode: mock.MagicMock,
        hasher: HasherChoice,
    ) -> None:
        """Test encoding error while dumping."""
        mock_b64encode.side_effect = ValueError

        signer = self.signer(encoder=B64URLEncoder, hasher=hasher)

        with pytest.raises(
                errors.EncodeError,
                match='data can not be encoded',
        ) as exc:
            self.sign(signer, self.data)
        assert isinstance(exc.value.__cause__, ValueError)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_dumps_invalid_compression_level(
        self,
        hasher: HasherChoice,
    ) -> None:
        """Test invalid compression level for dumps."""
        signer = self.signer(compressor=ZlibCompressor, hasher=hasher)

        with pytest.raises(
                errors.CompressionError,
                match='compression level must be between 1 and 9',
        ) as exc:
            self.sign(signer, self.data, compress=True, compression_level=10)
        assert exc.value.__cause__ is None

        with pytest.raises(
                errors.CompressionError,
                match='compression level must be between 1 and 9',
        ) as exc:
            # noinspection PyArgumentEqualDefault
            self.sign(signer, self.data, compress=True, compression_level=0)
        assert exc.value.__cause__ is None

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_compression_flag_non_ascii(self, hasher: HasherChoice) -> None:
        """Test error occurs when the compression flag is non-ascii."""
        with pytest.raises(
                errors.InvalidOptionError,
                match='compression flag character must be ASCII',
        ):
            self.signer(compression_flag=b'\x87', hasher=hasher)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_compression_flag_empty(self, hasher: HasherChoice) -> None:
        """Test error occurs when the compression flag is empty."""
        with pytest.raises(
                errors.InvalidOptionError,
                match='the compression flag character must have a value',
        ):
            self.signer(compression_flag=b'', hasher=hasher)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    @pytest.mark.parametrize(
        ('encoder', 'flag'),
        (
            (B64URLEncoder, b'A'),
            (B32Encoder, b'A'),
            (HexEncoder, b'A'),
        ),
    )
    def test_compression_flag_in_encoder_alphabet(
        self,
        encoder: typing.Type[EncoderInterface],
        flag: bytes,
        hasher: HasherChoice,
    ) -> None:
        """Test error occurs when the compression flag is in the encoder alphabet."""
        with pytest.raises(
                errors.InvalidOptionError,
                match='the compression flag character must not belong to the encoder',
        ):
            self.signer(compression_flag=flag, encoder=encoder, hasher=hasher)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_compression_ratio_out_of_bounds(self, hasher: HasherChoice) -> None:
        """Test error occurs when the compression ratio is out of bounds."""
        with pytest.raises(
                errors.InvalidOptionError,
                match='compression ratio must be',
        ):
            self.signer(compression_ratio=-1, hasher=hasher)

        with pytest.raises(
                errors.InvalidOptionError,
                match='compression ratio must be',
        ):
            self.signer(compression_ratio=100, hasher=hasher)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_load_file_error(self, hasher: HasherChoice) -> None:
        """Test error occurring during reading from a file."""
        file = mock.MagicMock()
        file.read.side_effect = TimeoutError
        signer = self.signer(hasher=hasher)

        with pytest.raises(
                errors.FileError,
                match='file can not be read',
        ):
            signer.load(file)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_dump_file_error(self, hasher: HasherChoice) -> None:
        """Test error occurring during writing to a file."""
        file = mock.MagicMock()
        file.write.side_effect = PermissionError
        signer = self.signer(hasher=hasher)

        with pytest.raises(
                errors.FileError,
                match='file can not be written',
        ):
            signer.dump(self.data, file)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    def test_dump_file_binary_conversion_error(self, hasher: HasherChoice) -> None:
        """Test error occurring during _write when file is in binary mode."""
        file = io.BytesIO()
        signer = self.signer(hasher=hasher)

        with pytest.raises(
                errors.ConversionError,
                match='can not be converted to bytes',
        ):
            signer._write(file, '\uD83D')


class TestsBlake2SerializerSignerTimestamp(
        TimestampSignerTestsBase,
        SerializerSignerTestsBase,
):
    """Blake2SerializerSigner tests with timestamp."""

    def signer(
        self,
        secret: typing.Union[None, str, bytes] = None,
        **kwargs: typing.Any,
    ) -> Signer:
        """Get the signer to test."""
        kwargs.setdefault('max_age', 5)
        return super().signer(secret, **kwargs)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    @pytest.mark.parametrize(
        'max_age',
        (None, 2, 2.5, timedelta(hours=2)),
    )
    @mock.patch('blake2signer.bases.time')
    def test_max_age_can_be_changed(
        self,
        mock_time: mock.MagicMock,
        max_age: typing.Union[None, int, float, timedelta],
        hasher: HasherChoice,
    ) -> None:
        """Test that max age can be changed correctly."""
        timestamp = int(time())
        mock_time.return_value = timestamp
        signer = self.signer(max_age=max_age, hasher=hasher)

        signed = self.sign(signer, self.data)
        assert self.data == self.unsign(signer, signed)

        if max_age:
            if isinstance(max_age, timedelta):
                mock_time.return_value += max_age.total_seconds()
            else:
                mock_time.return_value += max_age
            mock_time.return_value += 0.1  # It has to be a bit bigger than max_age

            with pytest.raises(
                    errors.ExpiredSignatureError,
                    match='signature has expired',
            ) as exc:
                self.unsign(signer, signed)
            assert exc.value.__cause__ is None
            assert exc.value.timestamp.timestamp() == timestamp

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        ('version', 'hasher', 'signed', 'compat'),
        (
            (
                '1.2.1',
                HasherChoice.blake2b,
                '3rXtVsLrCy3EP_6kdPW_NP6RaCXoIfwBa85o_w.H7LG0A.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                False,
            ),
            (
                '1.2.1',
                HasherChoice.blake2s,
                'cAJTWYO0t3gGIcPKYcxf1xcXYMDigA.H7LG0A.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                False,
            ),
            (
                '2.0.0',
                HasherChoice.blake2b,
                '4ghexnKJCLph0-yWvqWtYi5OKurbaK8sxfEe4g.H7LG0A.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
            (
                '2.0.0',
                HasherChoice.blake2s,
                'Cbf0LBzcJqgDi3Atd30Bxd7rJonn7Q.H7LG0A.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
            (
                '2.1.0',
                HasherChoice.blake2b,
                'WSJQGOWG5JYN0-5g3KD_dfni3mGf4IHdgjDtmg.H7LG0A.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
            (
                '2.1.0',
                HasherChoice.blake2s,
                '9txxz03E8LyX5XbrHpPSF73kGjIvwg.H7LG0A.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
            (
                '2.2.0',
                HasherChoice.blake2b,
                'BOTVypusKsx4ofDtwbAMbUMp-nfv1isosZe94g.H7LG0A.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
            (
                '2.2.0',
                HasherChoice.blake2s,
                'f9oZEMvwMkEbIfl56BjsPra3BeeF3Q.H7LG0A.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
            (
                '2.2.0',
                HasherChoice.blake3,
                '42nexVif5smuXxRqGre3PaT_DxzgBfCk7hHnyw.H7LG0A.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
            (
                '2.3.0',
                HasherChoice.blake2b,
                'PpBAgu92LbI4pBs9UEn3y6KaAtdz25iAzg54dg.H7LG0A.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
            (
                '2.3.0',
                HasherChoice.blake2s,
                'SWE7QcGvCBdqGYSRknvUXol06Nbggg.H7LG0A.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
            (
                '2.3.0',
                HasherChoice.blake3,
                'ez6GEqMnbBIIUYtdo3YJXEt6pmN_qQqAmZdeMg.H7LG0A.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
            (
                '2.4.0',
                HasherChoice.blake2b,
                'hZZkUGLgFFVWT-RA4JnHlt3t_nnkCdoU7wJVtg.H7LG0A.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
            (
                '2.4.0',
                HasherChoice.blake2s,
                'QKhqawqxageBFRF4ZmCoPrUAJjEWSw.H7LG0A.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
            (
                '2.4.0',
                HasherChoice.blake3,
                '10T9JBA_sRphLoSxIG6gA5oENMIAKqG_moBPFg.H7LG0A.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
        ),
    )
    def test_versions_compat(
        self,
        version: str,  # solely informational
        hasher: HasherChoice,
        signed: str,
        compat: bool,
    ) -> None:
        """Test if previous versions' signed data is compatible with the current one."""
        super().test_versions_compat(version, hasher, signed, compat)


class TestsBlake2SerializerSigner(SerializerSignerTestsBase):
    """Blake2SerializerSigner tests (without timestamp)."""

    def test_sign_unsign_with_different_signer(self) -> None:
        """Test signing and unsigning using different signer fails correctly."""
        signer1 = self.signer(max_age=None)  # Regular signer
        signer2 = self.signer(max_age=5)  # Timestamp signer

        self.mix_signers_sign_unsign(signer1, signer2)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        ('version', 'hasher', 'signed', 'compat'),
        (
            (
                '1.2.1',
                HasherChoice.blake2b,
                '8NMTFVqT3vaiLXsgYJFGJQuFV_sNtinMX_R3Wg.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
            (
                '1.2.1',
                HasherChoice.blake2s,
                'qMXQOTMMALCzgu7NSQ9lz1upgNwrEg.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
            (
                '2.0.0',
                HasherChoice.blake2b,
                'mGOPVc77_Wj4syM55kjkdTbUAJp8zq-P6Yov8w.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
            (
                '2.0.0',
                HasherChoice.blake2s,
                'qpTGzX5iLVj3KasTKJ38G8pUa17rJQ.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
            (
                '2.1.0',
                HasherChoice.blake2b,
                'CYd7Nkw038oKejApbHyNWJDTl9WnR5DEyQD9yQ.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
            (
                '2.1.0',
                HasherChoice.blake2s,
                'XnoeirEG8KD0FPJmBim0rzLuWQvOMQ.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
            (
                '2.2.0',
                HasherChoice.blake2b,
                'EFGMJB71JhnVh4g4EGNpAqsEBkrD8mC3bdBhug.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
            (
                '2.2.0',
                HasherChoice.blake2s,
                'goBSnffyopf866GSgdUDotfG71pOVA.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
            (
                '2.2.0',
                HasherChoice.blake3,
                '7MY6A3Nm8k2Sd9lTK61negSa52451DrtGb_y0g.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
            (
                '2.3.0',
                HasherChoice.blake2b,
                'X-eD0nscvIfGs3JgubIU9Ja8vbtzccSy0PVeug.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
            (
                '2.3.0',
                HasherChoice.blake2s,
                'vMyDuUnJar7pwjUtZ9noL1SCOwhzAg.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
            (
                '2.3.0',
                HasherChoice.blake3,
                '7U7dQhwqnOYNjSXruDBHpKwgbfiM8LoE4Q6oWw.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
            (
                '2.4.0',
                HasherChoice.blake2b,
                'Xvs2SOklzzZAKAfK22TNn1meysxuWj3aKWcKuQ.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
            (
                '2.4.0',
                HasherChoice.blake2s,
                'dWigMPiIE3ahrdn3yyBgvSpm9mqGqw.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
            (
                '2.4.0',
                HasherChoice.blake3,
                '-0EfMMvVRJwAIKbaxvUPVd05jQ55V6oamkhk-A.ImlzIGNvbXBhdCBlbnN1cmVkPyI',
                True,
            ),
        ),
    )
    def test_versions_compat(
        self,
        version: str,
        hasher: HasherChoice,
        signed: str,
        compat: bool,
    ) -> None:
        """Test if previous versions' signed data is compatible with the current one."""
        super().test_versions_compat(version, hasher, signed, compat)



"""Hashers handlers.

These are not meant to be used directly, but otherwise through a signer.
"""
import typing

from .blake3_package import blake3
from .blake3_package import has_blake3
from .blakehashers import BLAKE2Hasher
from .blakehashers import BLAKE3Hasher
from .blakehashers import HasherChoice

BLAKEHasher = typing.Union[BLAKE2Hasher, BLAKE3Hasher]

__all__ = (
    'blake3',  # Always import blake3 from this module
    'BLAKEHasher',
    'BLAKE2Hasher',
    'BLAKE3Hasher',
    'has_blake3',
    'HasherChoice',
)



"""Interfaces: abstract classes to define serializers, encoders and compressors."""

import typing
from abc import ABC
from abc import abstractmethod

from .errors import CompressionError


class SerializerInterface(ABC):
    """Serializer interface.

    Implement your own serializer inheriting from this class.
    """

    @abstractmethod
    def serialize(self, data: typing.Any, **kwargs: typing.Any) -> bytes:
        """Serialize given data.

        Args:
            data: Data to serialize.

        Keyword Args:
            **kwargs: Additional arguments for the serializer.

        Returns:
            Serialized data
        """

    @abstractmethod
    def unserialize(self, data: bytes, **kwargs: typing.Any) -> typing.Any:
        """Unserialize given serialized data.

        Args:
            data: Serialized data to unserialize.

        Keyword Args:
            **kwargs: Additional arguments for the serializer.

        Returns:
            Original data.
        """


class CompressorInterface(ABC):
    """Compressor interface.

    Implement your own compressor inheriting from this class.
    """

    @property
    @abstractmethod
    def default_compression_level(self) -> int:
        """Get the default compression level."""

    def scale_compression_level(self, level: int) -> int:
        """Scale the compression level to from 1 to 9 to a value for the compressor.

        Override this method if the compressor requires scaling the level.

        Args:
            level: Desired compression level from 1 to 9.

        Returns:
            Scaled compression level for the compressor.
        """
        return level

    def get_compression_level(self, level: typing.Optional[int]) -> int:
        """Return compression level for the compressor.

        It correctly converts the scale (if necessary), and the default value
        for None.

        Args:
            level: Desired compression level from 1 (least compressed) to 9 (most
                compressed) or None for the default.

        Returns:
            Correct compression level for the compressor.

        Raises:
            CompressionError: the compression level is out of bounds.
        """
        if level is None:
            return self.default_compression_level

        if level < 1 or level > 9:
            raise CompressionError('compression level must be between 1 and 9')

        return self.scale_compression_level(level)

    @abstractmethod
    def compress(self, data: bytes, *, level: int) -> bytes:
        """Compress given data.

        Args:
            data: Data to compress.

        Keyword Args:
            level: Desired compression level.

        Returns:
            Raw compressed data.
        """

    @abstractmethod
    def decompress(self, data: bytes) -> bytes:
        """Decompress given compressed data.

        Args:
            data: Compressed data to decompress.

        Returns:
            Original data.
        """


class EncoderInterface(ABC):
    """Encoder interface.

    Implement your own encoder inheriting from this class.

    Note:
        Verify that the encoder alphabet is ASCII (a check is enforced nevertheless).
    """

    @property
    @abstractmethod
    def alphabet(self) -> bytes:
        """Return the encoder alphabet characters.

        This is used to validate that separator characters and flags don't belong
        to this alphabet to prevent malfunctions.
        """

    @abstractmethod
    def encode(self, data: bytes) -> bytes:
        """Encode given data.

        Args:
            data: Data to encode.

        Returns:
            Encoded data.
        """

    @abstractmethod
    def decode(self, data: bytes) -> bytes:
        """Decode given encoded data.

        Args:
            data: Encoded data to decode.

        Returns:
            Original data.
        """



"""Signers tests."""

import typing

import pytest

from .bases import BaseTests
from .bases import Signature
from .bases import Signer
from .. import errors
from ..hashers import HasherChoice
from ..hashers import has_blake3
from ..signers import Blake2Signer


class TestsBlake2Signer(BaseTests):
    """Blake2Signer tests."""

    @property
    def signer_class(self) -> typing.Type[Signer]:
        """Get the signer class to test."""
        return Blake2Signer

    @property
    def signature_type(self) -> typing.Type[typing.Union[str, bytes]]:
        """Get the signature type for the signer (`bytes`)."""
        return bytes

    def sign(
        self,
        signer: Signer,
        data: typing.Any,
        **kwargs: typing.Any,
    ) -> typing.AnyStr:
        """Sign data with the signer."""
        return signer.sign(data)

    def unsign(
        self,
        signer: Signer,
        signed_data: typing.Union[str, bytes],
        **kwargs: typing.Any,
    ) -> typing.Any:
        """Unsign signed data with the signer."""
        return signer.unsign(signed_data)

    def sign_parts(
        self,
        signer: Signer,
        data: typing.Any,
        **kwargs: typing.Any,
    ) -> Signature:
        """Sign data with the signer in parts."""
        return signer.sign_parts(data)

    def unsign_parts(
        self,
        signer: Signer,
        signature: Signature,
        **kwargs: typing.Any,
    ) -> typing.Any:
        """Unsign signature with the signer."""
        return signer.unsign_parts(signature)

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        'hasher',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
            HasherChoice.blake3,
        ),
    )
    @pytest.mark.parametrize(
        'data',
        ('datadata', b'datadata'),
    )
    def test_data_can_be_string_or_bytes(
        self,
        data: typing.AnyStr,
        hasher: HasherChoice,
    ) -> None:
        """Test that data can be either bytes or string."""
        signer = self.signer(hasher=hasher)

        unsigned = self.unsign(signer, self.sign(signer, data))
        if isinstance(data, bytes):
            assert data == unsigned
        else:
            assert data.encode() == unsigned

    @pytest.mark.xfail(
        not has_blake3(),
        reason='blake3 is not installed',
        raises=errors.MissingDependencyError,
    )
    @pytest.mark.parametrize(
        ('version', 'hasher', 'signed', 'compat'),
        (
            (
                '1.2.1',
                HasherChoice.blake2b,
                '2exTNSllIBMpAkkd0SBv3r2_wew7q9t6SB_OVg.is compat ensured?',
                False,
            ),
            (
                '1.2.1',
                HasherChoice.blake2s,
                'nXsYsvLVNvqLatytrg_SevNvY_SLIw.is compat ensured?',
                False,
            ),
            (
                '2.0.0',
                HasherChoice.blake2b,
                '8lRrzthYNOHdxhkakSo34_jwOa_Mp_FQcFK8Sg.is compat ensured?',
                True,
            ),
            (
                '2.0.0',
                HasherChoice.blake2s,
                'L_jCeKUmWWAuXNATdnrvXmssf-IpGw.is compat ensured?',
                True,
            ),
            (
                '2.1.0',
                HasherChoice.blake2b,
                'Qu5_rlDNXDOO0ie9ZWbrkDgyxUrlQTmP-KJBLg.is compat ensured?',
                True,
            ),
            (
                '2.1.0',
                HasherChoice.blake2s,
                'TOsyPse3l0EWvhWPsQ6HAE30JlIYNA.is compat ensured?',
                True,
            ),
            (
                '2.2.0',
                HasherChoice.blake2b,
                '_RKzXIJZfg1cOV-mbNw9WX4wYBA2pcqdKrguoA.is compat ensured?',
                True,
            ),
            (
                '2.2.0',
                HasherChoice.blake2s,
                '8rsWR8Aoi8EeQI7CcS80zZ320Pr5Nw.is compat ensured?',
                True,
            ),
            (
                '2.2.0',
                HasherChoice.blake3,
                '9DbuBl_FqzvsspaCC6H23C8CEmRWH8mucCOXbA.is compat ensured?',
                True,
            ),
            (
                '2.3.0',
                HasherChoice.blake2b,
                'kHA3CsED0bqLXYZ0vGzuPbIU70lKeaUa-4PgAQ.is compat ensured?',
                True,
            ),
            (
                '2.3.0',
                HasherChoice.blake2s,
                'i_udo1_KBC9ff-iWqOoOHiq60YqojA.is compat ensured?',
                True,
            ),
            (
                '2.3.0',
                HasherChoice.blake3,
                '9OPOu3qaF616QQ81ut2bDtOf9O2Pqad_1lNZSw.is compat ensured?',
                True,
            ),
            (
                '2.4.0',
                HasherChoice.blake2b,
                'ZCLGwc0v0rzptgunF70vP1GrkNZDl4q7iGudCw.is compat ensured?',
                True,
            ),
            (
                '2.4.0',
                HasherChoice.blake2s,
                'PRdPO7n-rGAwLDVJ40jmbztPJYo6Ag.is compat ensured?',
                True,
            ),
            (
                '2.4.0',
                HasherChoice.blake3,
                'YoCHtjlT-s5m9yuJBn2DJix5vT5xMnuUVR5BBg.is compat ensured?',
                True,
            ),
        ),
    )
    def test_versions_compat(
        self,
        version: str,
        hasher: HasherChoice,
        signed: str,
        compat: bool,
    ) -> None:
        """Test if previous versions' signed data is compatible with the current one."""
        super().test_versions_compat(version, hasher, signed, compat)



"""Hashers tests."""

import typing
from abc import ABC
from abc import abstractmethod
from unittest import mock

import pytest

from .. import errors
from ..hashers import BLAKE2Hasher
from ..hashers import BLAKE3Hasher
from ..hashers import HasherChoice
from ..hashers import blake3

THasher = typing.TypeVar('THasher', BLAKE2Hasher, BLAKE3Hasher)


class BaseTests(typing.Generic[THasher], ABC):
    """Base class for the hashers' tests."""
    secrets = (b'averysecretsecret', b'0123456789012345')
    person = b'acab'
    digest_size = 16

    @property
    @abstractmethod
    def hasher_class(self) -> typing.Type[THasher]:
        """Get the hasher class to test."""

    @property
    @abstractmethod
    def hasher_choice(self) -> HasherChoice:
        """Get the default hasher choice."""

    def get_hasher(
        self,
        hasher: typing.Optional[HasherChoice] = None,
        *,
        secrets: typing.Optional[typing.Tuple[bytes]] = None,
        digest_size: typing.Optional[int] = None,
        person: typing.Optional[bytes] = None,
    ) -> THasher:
        """Get the hasher instance to test."""
        return self.hasher_class(
            hasher or self.hasher_choice,
            secrets=secrets or self.secrets,
            digest_size=digest_size or self.digest_size,
            person=person or self.person,
        )


class TestsBLAKE2Hasher(BaseTests[BLAKE2Hasher]):
    """BLAKE2Hasher tests."""

    @property
    def hasher_class(self) -> typing.Type[BLAKE2Hasher]:
        """Get the hasher class to test."""
        return BLAKE2Hasher

    @property
    def hasher_choice(self) -> HasherChoice:
        """Get the hasher class to test."""
        return HasherChoice.blake2b

    @pytest.mark.parametrize(
        'choice',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
        ),
    )
    def test_validate_digest_size_too_long(self, choice: HasherChoice) -> None:
        """Test that digest_size is being correctly validated."""
        with pytest.raises(
                errors.InvalidOptionError,
                match='digest_size should be lower',
        ):
            self.get_hasher(choice, digest_size=128)

    @pytest.mark.parametrize(
        'choice',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
        ),
    )
    def test_derive_person(self, choice: HasherChoice) -> None:
        """Test that person is correctly derived."""
        hasher = self.get_hasher(choice)

        assert self.person != hasher._person
        assert hasher._hasher.PERSON_SIZE == len(hasher._person)

    @pytest.mark.parametrize(
        'choice',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
        ),
    )
    def test_derive_secret(self, choice: HasherChoice) -> None:
        """Test that secrets are correctly derived."""
        hasher = self.get_hasher(choice)

        assert len(self.secrets) == len(hasher.keys)
        assert self.secrets != hasher.keys
        assert sorted(self.secrets) != sorted(hasher.keys)

        for key in hasher.keys:
            assert hasher._hasher.MAX_KEY_SIZE == len(key)

        # Ensure keys are actually derived
        with mock.patch.object(self.hasher_class, '_derive_key') as mock_derive_key:
            hasher = self.get_hasher(choice)
        calls = []
        for secret in self.secrets:
            calls.append(mock.call(secret, person=hasher._person))
        mock_derive_key.assert_has_calls(calls)

        # Ensure the hasher gets called properly
        with mock.patch(f'blake2signer.hashers.blakehashers.hashlib.{choice}') as mock_hasher:
            mock_hasher.PERSON_SIZE = 8
            mock_hasher.MAX_KEY_SIZE = 12
            mock_hasher.MAX_DIGEST_SIZE = 16
            mock_hasher.return_value.digest.return_value = b'abc123'

            self.get_hasher(choice)

        calls = [mock.call(self.person, digest_size=8), mock.call().digest()]
        for secret in self.secrets:
            calls.append(mock.call(secret, digest_size=12, person=b'abc123'))
            calls.append(mock.call().digest())
        mock_hasher.assert_has_calls(calls)

    @pytest.mark.parametrize(
        ('choice', 'data', 'salt', 'expected_digest'),
        (
            (
                HasherChoice.blake2b,
                b'datadata',
                b'',
                b'I\xb5v\x9d\xd7"\xec\xd8Q\x0c\x94\xd0\xa1\x02z?',
            ),
            (
                HasherChoice.blake2b,
                b'datadata',
                b'salt',
                b'\xb8\xa4\x84o\xbf\xf6>\xc1\x9dLN/\xbe\\\xd1>',
            ),
            (
                HasherChoice.blake2s,
                b'datadata',
                b'',
                b"\xf9`<\x8c\x85\xe8o+\x07'\x84s=\xc5'\x8a",
            ),
            (
                HasherChoice.blake2s,
                b'datadata',
                b'salt',
                b'\xfb,T\xc0\xc5\xf5\xf4\xf9\xfc\x8d\xb1T\xceY\x03\xdb',
            ),
        ),
    )
    def test_digest(
        self,
        choice: HasherChoice,
        data: bytes,
        salt: bytes,
        expected_digest: bytes,
    ) -> None:
        """Test that digest is correctly calculated."""
        hasher = self.get_hasher(choice)

        digest = hasher.digest(b'datadata', key=hasher.signing_key, salt=salt)

        assert expected_digest == digest

    @pytest.mark.parametrize(
        'choice',
        (
            HasherChoice.blake2b,
            HasherChoice.blake2s,
        ),
    )
    def test_signing_key_is_the_newest_one(self, choice: HasherChoice) -> None:
        """Test that the signing key is the newest one."""
        hasher = self.get_hasher(choice)

        assert hasher.keys[-1] == hasher.signing_key


class TestsBLAKE2HasherPy38(TestsBLAKE2Hasher):
    """BLAKE2Hasher tests, hack for Pytest under Python <=3.8."""

    __new__ = object.__new__


class TestsBLAKE3Hasher(BaseTests[BLAKE3Hasher]):
    """BLAKE3Hasher tests."""

    @property
    def hasher_class(self) -> typing.Type[BLAKE3Hasher]:
        """Get the hasher class to test."""
        return BLAKE3Hasher

    @property
    def hasher_choice(self) -> HasherChoice:
        """Get the hasher class to test."""
        return HasherChoice.blake3

    def test_validate_digest_size_unlimited(self) -> None:
        """Test that blake3 has no maximum digest size."""
        hasher = self.get_hasher(digest_size=1_000)

        assert hasher.max_digest_size is None

    def test_derive_person(self) -> None:
        """Test that person is not derived."""
        hasher = self.get_hasher()

        assert self.person == hasher._person

    def test_derive_secret(self) -> None:
        """Test that secret is correctly derived."""
        hasher = self.get_hasher()

        assert len(self.secrets) == len(hasher.keys)
        assert self.secrets != hasher.keys
        assert sorted(self.secrets) != sorted(hasher.keys)

        for key in hasher.keys:
            assert blake3().key_size == len(key)

        # Ensure keys are actually derived
        with mock.patch.object(self.hasher_class, '_derive_key') as mock_derive_key:
            hasher = self.get_hasher()
        calls = []
        for secret in self.secrets:
            calls.append(mock.call(secret, person=hasher._person))
        mock_derive_key.assert_has_calls(calls)

        # Ensure the hasher gets called properly
        with mock.patch('blake2signer.hashers.blakehashers.blake3') as mock_hasher:
            mock_hasher.return_value.key_size = 16

            self.get_hasher()

        calls = []  # person is not derived
        derive_key_context = 'blake2signer 2021-12-29 18:04:37 BLAKE3Hasher key derivation'
        for secret in self.secrets:
            calls.append(mock.call(secret, derive_key_context=derive_key_context))
            calls.append(mock.call().update(self.person))
            calls.append(mock.call().digest(length=16))
        mock_hasher.assert_has_calls(calls)

    @pytest.mark.parametrize(
        ('data', 'salt', 'expected_digest'),
        (
            (b'datadata', b'', b"\xfe\x88\x96\xe5\xdf:\xb6f\xe1}\x9f\x18\x9a'\xad\x8d"),
            (b'datadata', b'salt', b'\xeez6\xf20\xca\xcae\x93\xc3\xcchh\x90(\xdf'),
        ),
    )
    def test_digest(
        self,
        data: bytes,
        salt: bytes,
        expected_digest: bytes,
    ) -> None:
        """Test that digest is correctly calculated."""
        hasher = self.get_hasher()

        digest = hasher.digest(b'datadata', key=hasher.signing_key, salt=salt)

        assert expected_digest == digest

    def test_signing_key_is_the_newest_one(self) -> None:
        """Test that the signing key is the newest one."""
        hasher = self.get_hasher()

        assert hasher.keys[-1] == hasher.signing_key


class TestsBLAKE3HasherPy38(TestsBLAKE3Hasher):
    """BLAKE3Hasher tests, hack for Pytest under Python <=3.8."""

    __new__ = object.__new__



"""Encoders: classes that implement the EncoderInterface."""

from .interfaces import EncoderInterface
from .utils import b32decode
from .utils import b32encode
from .utils import b64decode
from .utils import b64encode
from .utils import hexdecode
from .utils import hexencode


class B64URLEncoder(EncoderInterface):
    """Base64 URL safe encoder."""

    @property
    def alphabet(self) -> bytes:
        """Return the encoder alphabet characters."""
        return b'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_-'

    def encode(self, data: bytes) -> bytes:
        """Encode given data to base64 URL safe without padding.

        Args:
            data: Data to encode.

        Returns:
            Encoded data.
        """
        return b64encode(data)

    def decode(self, data: bytes) -> bytes:
        """Decode given encoded data from base64 URL safe without padding.

        Args:
            data: Data to decode.

        Returns:
            Original data.
        """
        return b64decode(data)


class B32Encoder(EncoderInterface):
    """Base32 encoder."""

    @property
    def alphabet(self) -> bytes:
        """Return the encoder alphabet characters."""
        return b'ABCDEFGHIJKLMNOPQRSTUVWXYZ234567'

    def encode(self, data: bytes) -> bytes:
        """Encode given data to base32 without padding.

        Args:
            data: Data to encode.

        Returns:
            Encoded data.
        """
        return b32encode(data)

    def decode(self, data: bytes) -> bytes:
        """Decode given encoded data from base32 without padding.

        Args:
            data: Data to decode.

        Returns:
            Original data.
        """
        return b32decode(data)


class HexEncoder(EncoderInterface):
    """Hexadecimal encoder."""

    @property
    def alphabet(self) -> bytes:
        """Return the encoder alphabet characters."""
        return b'ABCDEF0123456789'

    def encode(self, data: bytes) -> bytes:
        """Encode given data to hexadecimal.

        Args:
            data: Data to encode.

        Returns:
            Encoded data.
        """
        return hexencode(data)

    def decode(self, data: bytes) -> bytes:
        """Decode given encoded data from hexadecimal.

        Args:
            data: Data to decode.

        Returns:
            Original data.
        """
        return hexdecode(data)



"""Errors: contains all errors and exceptions raised by this lib.

Note:
    Here's the hierarchy tree:

        SignerError
            |
            |-- InvalidOptionError: given option value is out of bounds, has the wrong
            |                       format or type.
            |
            |-- MissingDependencyError: a required dependency is not installed.
            |
            |
            |-- DataError: generic data error.
                    |
                    |-- SignedDataError: error that occurred for *signed data*.
                    |       |
                    |       |-- SignatureError: error encountered while dealing with
                    |       |       |           the signature.
                    |       |       |
                    |       |       |-- InvalidSignatureError: the signature is not
                    |       |               |                  valid.
                    |       |               |
                    |       |               |-- ExpiredSignatureError: the signature
                    |       |                                          has expired.
                    |       |
                    |       |-- UnserializationError: given data could not be
                    |       |                         unserialized.
                    |       |
                    |       |-- DecompressionError: given data could not be decompressed.
                    |       |
                    |       |-- DecodeError: given data could not be decoded.
                    |       |
                    |       |-- ConversionError: given data could not be converted
                    |       |                    to bytes.
                    |       |
                    |       |-- FileError: error while reading the file.
                    |
                    |-- UnsignedDataError: error that occurred for *data to be signed*.
                            |
                            |-- SerializationError: given data could not be serialized.
                            |
                            |-- CompressionError: given data could not be compressed.
                            |
                            |-- EncodeError: given data could not be encoded.
                            |
                            |-- ConversionError: given data could not be converted
                            |                    to bytes.
                            |
                            |-- FileError: error while writing the file.
"""

import typing
from datetime import datetime


class SignerError(Exception):
    """Base exception for all errors."""


class InvalidOptionError(SignerError):
    """Invalid option error.

    Means that given value is out of bounds or has the wrong format or type for
    the option.
    """


class MissingDependencyError(SignerError):
    """Missing dependency error.

    Means that a required dependency is not installed.
    """


class DataError(SignerError):
    """Data error.

    Generic data error meaning that given data could not be processed correctly.

    All exceptions regarding data handling depends on this one, so you can safely
    catch it to deal with data errors (both signed and to be signed).
    """


class SignedDataError(DataError):
    """Signed data error.

    Generic data error that occurred for signed data that is being processed.

    All exceptions regarding signed data handling depends on this one, so you can
    safely catch it to deal with signed data errors (produced during `unsign`,
    `unsign_parts`, `loads`, `loads_parts`, or `load`).
    """


class UnsignedDataError(DataError):
    """Unsigned data error.

    Generic data error that occurred for data to be signed that is being processed.

    All exceptions regarding non-signed data handling depends on this one, so you
    can safely catch it to deal with non-signed data errors (produced during
    `sign`, `sign_parts`, `dumps`, `dumps_parts` or `dump`).
    """


class SignatureError(SignedDataError):
    """Signature error.

    Means that an error was encountered while dealing with some part of the
    signature.
    """


class InvalidSignatureError(SignatureError):
    """Invalid signature error.

    Means that the signature is not valid.
    """


class ExpiredSignatureError(InvalidSignatureError):
    """Expired signature error.

    Means that the signature has expired.
    """

    # ToDo: D417 is a false positive, see https://github.com/PyCQA/pydocstyle/issues/514
    def __init__(self, *args: typing.Any, timestamp: datetime) -> None:  # noqa: D417
        """Initialize self.

        Args:
            *args: Additional positional arguments, see `Exception.__init__`.

        Keyword Args:
            timestamp: An aware datetime object indicating when the signature was done.
        """
        super().__init__(*args)

        self.timestamp: datetime = timestamp


class UnserializationError(SignedDataError):
    """Unserialization error.

    Means that given data could not be unserialized.
    """


class SerializationError(UnsignedDataError):
    """Serialization error.

    Means that given data could not be serialized.
    """


class DecompressionError(SignedDataError):
    """Decompression error.

    Means that given data could not be decompressed.
    """


class CompressionError(UnsignedDataError):
    """Compression error.

    Means that given data could not be compressed.
    """


class DecodeError(SignedDataError):
    """Decode error.

    Means that given data could not be decoded.
    """


class EncodeError(UnsignedDataError):
    """Encode error.

    Means that given data could not be encoded.
    """


class ConversionError(SignedDataError, UnsignedDataError):
    """Conversion error.

    Means that given data could not be converted to bytes. This can happen for
    any process.
    """


class FileError(SignedDataError, UnsignedDataError):
    """File error.

    Means that an operation pertaining a file failed. This can happen during
    `dump` or `load`.
    """



"""Bases: base classes for signers."""

import os
import typing
from abc import ABC
from abc import abstractmethod
from datetime import timedelta
from secrets import compare_digest
from time import time

from .encoders import B64URLEncoder
from .errors import ExpiredSignatureError
from .errors import FileError
from .errors import InvalidOptionError
from .errors import InvalidSignatureError
from .errors import SignatureError
from .hashers import BLAKE2Hasher
from .hashers import BLAKE3Hasher
from .hashers import BLAKEHasher
from .hashers import HasherChoice
from .interfaces import EncoderInterface
from .mixins import EncoderMixin
from .mixins import Mixin
from .utils import file_mode_is_text
from .utils import ordinal
from .utils import timestamp_to_aware_datetime


class SignedDataParts(typing.NamedTuple):
    """Parts of a signed data container."""

    data: bytes
    salt: bytes
    signature: bytes


class TimestampedDataParts(typing.NamedTuple):
    """Parts of a timestamped data container."""

    data: bytes
    timestamp: int


class Blake2SignatureDump(typing.NamedTuple):
    """Signature container."""

    signature: str  # Composite signature
    data: str


class Blake2Signature(typing.NamedTuple):
    """Signature container."""

    signature: bytes  # Composite signature
    data: bytes


Secret = typing.Union[str, bytes]


class Base(Mixin, ABC):
    """Base class containing the minimum for a signer."""

    Hashers = HasherChoice  # Sugar to avoid having to import the enum

    MIN_SECRET_SIZE: int = 16
    """Minimum secret size allowed (during instantiation)."""

    MIN_DIGEST_SIZE: int = 16
    """Minimum digest size allowed (during instantiation)."""

    DEFAULT_DIGEST_SIZE: int = 16  # 16 bytes is good security/size tradeoff
    """Default digest size to use when no digest size is indicated."""

    def __init__(
        self,
        secret: typing.Union[Secret, typing.Sequence[Secret]],
        *,
        personalisation: typing.Union[str, bytes] = b'',
        digest_size: typing.Optional[int] = None,
        hasher: typing.Union[HasherChoice, str] = HasherChoice.blake2b,
        deterministic: bool = False,
        separator: typing.Union[str, bytes] = b'.',
    ) -> None:
        """Sign and verify signed data using BLAKE in keyed hashing mode.

        Args:
            secret: Secret value which will be derived using BLAKE to produce the
                signing key. The minimum secret size is enforced to 16 bytes and there
                is no maximum. You can optionally provide a sequence of secrets, oldest
                to newest, that are used during signature check to allow for secret
                rotation. The last, newest, secret is used for signing.

        Keyword Args:
            personalisation (optional): Personalisation string to force the hash
                function to produce different digests for the same input. It is
                derived using BLAKE to ensure it fits the hasher limits, so it
                has no practical size limit. It defaults to the class name.
            digest_size (optional): Size of output signature (digest) in bytes
                (defaults to 16 bytes). The minimum size is enforced to 16 bytes.
            hasher (optional): Hash function to use: blake2b (default), blake2s
                or blake3.
            deterministic (optional): Define if signatures are deterministic or
                non-deterministic (default). Non-deterministic sigs are preferred,
                and achieved through the use of a random salt. For deterministic
                sigs, no salt is used: this means that for the same payload, the
                same sig is obtained (the advantage is that the sig is shorter).
            separator (optional): Character to separate the signature and the
                payload. It must not belong to the encoder alphabet and be ASCII
                (defaults to ".").

        Raises:
            ConversionError: A bytes parameter is not bytes and can't be converted
                to bytes.
            InvalidOptionError: A parameter is out of bounds.
        """
        hasher_choice = self._validate_hasher_choice(hasher)
        digest_size = self._validate_digest_size(digest_size)
        separator = self._validate_separator(separator)
        person = self._validate_person(personalisation)
        secrets = self._validate_secret(secret)

        if deterministic:
            person += b'Deterministic'
        person += self.__class__.__name__.encode()

        self._deterministic: bool = deterministic
        self._separator: bytes = separator

        self._hasher = self._get_hasher(
            hasher_choice,
            secrets=secrets,
            digest_size=digest_size,
            person=person,
        )

    def _validate_secret(
        self,
        secret: typing.Union[Secret, typing.Sequence[Secret]],
    ) -> typing.Tuple[bytes, ...]:
        """Validate the secret value and return it clean.

        Args:
            secret: Secret value to validate.

        Returns:
            Cleaned secrets tuple.

        Raises:
            ConversionError: The value is not bytes and can't be converted to bytes.
            InvalidOptionError: The value is out of bounds.
        """

        def coerce(secret_: Secret, position_: int) -> bytes:
            """Coerce a given secret into bytes.

            Args:
                secret_: Secret value to validate.
                position_: Position of the secret in the sequence.

            Returns:
                A secret value as bytes.

            Raises:
                InvalidOptionError: the secret in given position is invalid.
            """
            coerced = self._force_bytes(secret_)
            if len(coerced) < self.MIN_SECRET_SIZE:
                raise InvalidOptionError(
                    f'the {ordinal(position_)} secret should be longer than '
                    + f'{self.MIN_SECRET_SIZE} bytes',
                )

            return coerced

        secrets: typing.Sequence[Secret]
        if isinstance(secret, (str, bytes)):
            secrets = [secret]
        else:
            secrets = secret

        return tuple(coerce(dirty, position) for position, dirty in enumerate(secrets, start=1))

    def _validate_person(self, person: typing.Union[str, bytes]) -> bytes:
        """Validate the personalisation value and return it clean.

        Args:
            person: Personalisation value to validate.

        Returns:
            Cleaned personalisation value.

        Raises:
            ConversionError: The value is not bytes and can't be converted to bytes.
        """
        return self._force_bytes(person)

    def _validate_digest_size(self, digest_size: typing.Optional[int]) -> int:
        """Validate the digest_size value and return it clean.

        Args:
            digest_size: Digest size value to validate.

        Returns:
            Cleaned digest size value.

        Raises:
            InvalidOptionError: The value is out of bounds.
        """
        if digest_size is None:
            digest_size = self.DEFAULT_DIGEST_SIZE

        if digest_size < self.MIN_DIGEST_SIZE:
            raise InvalidOptionError(
                f'digest_size should be bigger than or equal to {self.MIN_DIGEST_SIZE}',
            )

        return digest_size

    @staticmethod
    def _validate_hasher_choice(hasher: typing.Union[HasherChoice, str]) -> HasherChoice:
        """Validate the hasher choice.

        Args:
            hasher: the hasher choice to validate.

        Returns:
            A validated hasher choice as HasherChoice.

        Raises:
            InvalidOptionError: Invalid hasher choice.
        """
        try:
            choice = HasherChoice(hasher)
        except ValueError:
            raise InvalidOptionError(
                f'invalid hasher choice, must be one of: {", ".join(h for h in HasherChoice)}',
            )

        return choice

    def _validate_separator(self, separator: typing.Union[str, bytes]) -> bytes:
        """Validate the separator value and return it clean.

        Args:
            separator: Separator value to validate.

        Returns:
            Cleaned separator value.

        Raises:
            ConversionError: The value is not bytes and can't be converted to bytes.
            InvalidOptionError:  The value is out of bounds.
        """
        if not separator:
            raise InvalidOptionError('the separator character must have a value')

        if not separator.isascii():
            raise InvalidOptionError('the separator character must be ASCII')

        return self._force_bytes(separator)

    @staticmethod
    def _get_hasher(
        hasher: HasherChoice,
        *,
        secrets: typing.Tuple[bytes, ...],
        digest_size: int,
        person: bytes,
    ) -> BLAKEHasher:
        """Get the proper hasher instance regarding the choice."""
        hasher_class: typing.Type[BLAKEHasher]

        if hasher in {HasherChoice.blake2b, HasherChoice.blake2s}:
            hasher_class = BLAKE2Hasher
        else:
            hasher_class = BLAKE3Hasher

        return hasher_class(
            hasher,
            secrets=secrets,
            digest_size=digest_size,
            person=person,
        )


class Blake2SignerBase(EncoderMixin, Base, ABC):
    """Base class for a signer based on BLAKE in keyed hashing mode."""

    def __init__(
        self,
        secret: typing.Union[Secret, typing.Sequence[Secret]],
        *,
        personalisation: typing.Union[str, bytes] = b'',
        digest_size: typing.Optional[int] = None,
        hasher: typing.Union[HasherChoice, str] = HasherChoice.blake2b,
        deterministic: bool = False,
        separator: typing.Union[str, bytes] = b'.',
        encoder: typing.Type[EncoderInterface] = B64URLEncoder,
    ) -> None:
        """Sign and verify signed data using BLAKE in keyed hashing mode.

        Args:
            secret: Secret value which will be derived using BLAKE to produce the
                signing key. The minimum secret size is enforced to 16 bytes and there
                is no maximum. You can optionally provide a sequence of secrets, oldest
                to newest, that are used during signature check to allow for secret
                rotation. The last, newest, secret is used for signing.

        Keyword Args:
            personalisation (optional): Personalisation string to force the hash
                function to produce different digests for the same input. It is
                derived using BLAKE to ensure it fits the hasher limits, so it
                has no practical size limit. It defaults to the class name.
            digest_size (optional): Size of output signature (digest) in bytes
                (defaults to 16 bytes). The minimum size is enforced to 16 bytes.
            hasher (optional): Hash function to use: blake2b (default), blake2s
                or blake3.
            deterministic (optional): Define if signatures are deterministic or
                non-deterministic (default). Non-deterministic sigs are preferred,
                and achieved through the use of a random salt. For deterministic
                sigs, no salt is used: this means that for the same payload, the
                same sig is obtained (the advantage is that the sig is shorter).
            separator (optional): Character to separate the signature and the
                payload. It must not belong to the encoder alphabet and be ASCII
                (defaults to ".").
            encoder (optional): Encoder class to use for the signature, nothing
                else is encoded (defaults to a Base64 URL safe encoder).

        Raises:
            ConversionError: A bytes parameter is not bytes and can't be converted
                to bytes.
            InvalidOptionError: A parameter is out of bounds.
        """
        super().__init__(
            secret,
            personalisation=personalisation,
            digest_size=digest_size,
            hasher=hasher,
            separator=separator,
            deterministic=deterministic,
            encoder=encoder,
        )

    def _validate_separator(self, separator: typing.Union[str, bytes]) -> bytes:
        """Validate the separator value and return it clean.

        Args:
            separator: Separator value to validate.

        Returns:
            Cleaned separator value.

        Raises:
            ConversionError: The value is not bytes and can't be converted to bytes.
            InvalidOptionError: The value is out of bounds.
        """
        sep = super()._validate_separator(separator)

        if sep in self._encoder.alphabet:
            raise InvalidOptionError(
                'the separator character must not belong to the encoder alphabet',
            )

        return sep

    def _get_salt(self) -> bytes:
        """Get a salt for the signature considering its type.

        For non-deterministic signatures, a pseudo random salt is generated.

        Returns:
            A pseudorandom salt for the signature if non-deterministic, otherwise an
            empty bytes string.
        """
        if self._deterministic:
            return b''

        salt_size = self._hasher.salt_size
        salt = os.urandom(salt_size)
        # Produce an encoded salt to use it as is, so we don't have to deal with
        # decoding it when unsigning. The only downside is that we loose a few
        # bits, but it's tolerable since we are using the maximum allowed size.
        return self._encode(salt)[:salt_size]

    def _force_bytes_parts(
        self,
        signature: typing.Union[Blake2Signature, Blake2SignatureDump],
    ) -> Blake2Signature:
        """Force given value into bytes, meaning a Blake2Signature container."""
        return Blake2Signature(
            data=self._force_bytes(signature.data),
            signature=self._force_bytes(signature.signature),
        )

    def _compose(self, data: bytes, *, signature: bytes) -> bytes:
        """Compose data and signature into a single stream."""
        return signature + self._separator + data

    def _decompose(self, signed_data: bytes) -> SignedDataParts:
        """Decompose a signed data stream into its parts.

        Args:
            signed_data: the signed data stream to decompose.

        Returns:
            The decomposed signed data in a container.

        Raises:
            SignatureError: Invalid signed data.
        """
        if self._separator not in signed_data:
            raise SignatureError('separator not found in signed data')

        composite_signature, data = signed_data.split(self._separator, 1)

        if not composite_signature:
            raise SignatureError('signature information is missing')

        if self._deterministic:
            salt = b''
            signature = composite_signature
        else:
            salt_size = self._hasher.salt_size
            salt = composite_signature[:salt_size]
            signature = composite_signature[salt_size:]

        return SignedDataParts(data=data, salt=salt, signature=signature)

    def _signify(self, *, data: bytes, salt: bytes, key: bytes) -> bytes:
        """Return signature for given data using salt and all the hasher options.

        The signature is encoded using the chosen encoder.

        Keyword Args:
            data: data to sign.
            salt: salt to use in the signature.
            key: the signing key.

        Returns:
            An encoded signature.
        """
        signature = self._hasher.digest(data, key=key, salt=salt)

        return self._encode(signature)

    def _sign(self, data: bytes) -> bytes:
        """Sign given data and produce a signature stream composed of salt and signature.

        The signature stream (salt and signature) is encoded using the chosen encoder.

        Args:
            data: data to sign.

        Returns:
            A signature stream containing salt and signature.
        """
        salt = self._get_salt()
        signature = self._signify(data=data, salt=salt, key=self._hasher.signing_key)

        return salt + signature

    def _unsign(self, parts: SignedDataParts) -> bytes:
        """Verify signed data parts and recover original data.

        Args:
            parts: Signed data parts to unsign.

        Returns:
            Original data.

        Raises:
            InvalidSignatureError: Signed data signature is invalid.
        """
        for key in reversed(self._hasher.keys):
            signature = self._signify(data=parts.data, salt=parts.salt, key=key)
            if compare_digest(signature, parts.signature):
                return parts.data

        raise InvalidSignatureError('signature is not valid')


class Blake2TimestampSignerBase(Blake2SignerBase, ABC):
    """Base class for a timestamp signer based on BLAKE in keyed hashing mode."""

    def _get_timestamp(self) -> bytes:
        """Get the encoded timestamp value."""
        timestamp = int(time())  # It's easier to encode and decode an integer
        try:
            timestamp_b = timestamp.to_bytes(4, 'big', signed=False)
        except OverflowError:  # This will happen in ~2106-02-07
            raise RuntimeError(
                'can not represent this timestamp in bytes: this library is '
                + 'too old and needs to be updated!',
            )

        return self._encode(timestamp_b)

    def _decode_timestamp(self, encoded_timestamp: bytes) -> int:
        """Decode an encoded timestamp whose signature should have been validated.

        Args:
            encoded_timestamp: encoded timestamp to decode.

        Returns:
            An integer timestamp.

        Raises:
            DecodeError: Timestamp can't be decoded.
        """
        return int.from_bytes(self._decode(encoded_timestamp), 'big', signed=False)

    def _compose_timestamp(self, data: bytes, *, timestamp: bytes) -> bytes:
        """Compose timestamp value with data."""
        return timestamp + self._separator + data

    def _decompose_timestamp(self, timestamped_data: bytes) -> TimestampedDataParts:
        """Decompose data + timestamp value.

        Args:
            timestamped_data: composed data with timestamp to decompose.

        Returns:
            A container with data and timestamp.

        Raises:
            SignatureError: Invalid timestamped data.
            DecodeError: Timestamp can't be decoded.
        """
        if self._separator not in timestamped_data:
            raise SignatureError('separator not found in timestamped data')

        encoded_timestamp, data = timestamped_data.split(self._separator, 1)

        if not encoded_timestamp:
            raise SignatureError('timestamp information is missing')

        timestamp = self._decode_timestamp(encoded_timestamp)

        return TimestampedDataParts(data=data, timestamp=timestamp)

    @staticmethod
    def _get_ttl_from_max_age(max_age: typing.Union[int, float, timedelta]) -> float:
        """Get the time-to-live value in seconds."""
        if isinstance(max_age, timedelta):
            return max_age.total_seconds()

        return float(max_age)

    def _sign_with_timestamp(self, data: bytes) -> bytes:
        """Sign given data and produce a timestamped signature stream.

        The timestamped signature stream (timestamp, signature and salt) is
        encoded using the chosen encoder.

        Args:
            data: data to sign.

        Returns:
            A signature stream composed of salt, signature and timestamp.
        """
        timestamp = self._get_timestamp()
        timestamped_data = self._compose_timestamp(data, timestamp=timestamp)

        return self._compose(timestamp, signature=self._sign(timestamped_data))

    def _unsign_with_timestamp(
        self,
        parts: SignedDataParts,
        *,
        max_age: typing.Union[None, int, float, timedelta],
    ) -> bytes:
        """Verify signed data parts with timestamp and recover original data.

        If `max_age` is not provided, then the timestamp is not checked (the
        signature is always checked).

        Args:
            parts: Signed data parts to unsign.

        Keyword Args:
            max_age (optional): Ensure the signature is not older than this time
                in seconds.

        Returns:
            Original data.

        Raises:
            SignatureError: Signed data structure is not valid.
            InvalidSignatureError: Signed data signature is invalid.
            ExpiredSignatureError: Signed data signature has expired.
            DecodeError: Timestamp can't be decoded.
        """
        timestamped_data = self._unsign(parts)

        timestamped_parts = self._decompose_timestamp(timestamped_data)

        if max_age is None:
            return timestamped_parts.data

        now = time()
        age = now - timestamped_parts.timestamp
        ttl = self._get_ttl_from_max_age(max_age)

        if age > ttl:
            raise ExpiredSignatureError(
                f'signature has expired, age {age} > {ttl} seconds',
                timestamp=timestamp_to_aware_datetime(timestamped_parts.timestamp),
            )

        if age < 0:  # Signed in the future
            raise ExpiredSignatureError(
                f'signature has expired, age {age} < 0 seconds',
                timestamp=timestamp_to_aware_datetime(timestamped_parts.timestamp),
            )

        return timestamped_parts.data


class Blake2DualSignerBase(Blake2TimestampSignerBase, ABC):
    """Base class for a dual signer: with and without timestamp."""

    def __init__(
        self,
        secret: typing.Union[Secret, typing.Sequence[Secret]],
        *,
        max_age: typing.Union[None, int, float, timedelta] = None,
        personalisation: typing.Union[str, bytes] = b'',
        digest_size: typing.Optional[int] = None,
        hasher: typing.Union[HasherChoice, str] = HasherChoice.blake2b,
        deterministic: bool = False,
        separator: typing.Union[str, bytes] = b'.',
        encoder: typing.Type[EncoderInterface] = B64URLEncoder,
    ) -> None:
        """Sign and verify signed and optionally timestamped data using BLAKE.

        It uses BLAKE in keyed hashing mode.

        Setting `max_age` will produce a timestamped signed stream.

        Args:
            secret: Secret value which will be derived using BLAKE to produce the
                signing key. The minimum secret size is enforced to 16 bytes and there
                is no maximum. You can optionally provide a sequence of secrets, oldest
                to newest, that are used during signature check to allow for secret
                rotation. The last, newest, secret is used for signing.

        Keyword Args:
            max_age (optional): Use a timestamp signer instead of a regular one
                to ensure that the signature is not older than this time in seconds.
            personalisation (optional): Personalisation string to force the hash
                function to produce different digests for the same input. It is
                derived using BLAKE to ensure it fits the hasher limits, so it
                has no practical size limit. It defaults to the class name.
            digest_size (optional): Size of output signature (digest) in bytes
                (defaults to 16 bytes). The minimum size is enforced to 16 bytes.
            hasher (optional): Hash function to use: blake2b (default), blake2s
                or blake3.
            deterministic (optional): Define if signatures are deterministic or
                non-deterministic (default). Non-deterministic sigs are preferred,
                and achieved through the use of a random salt. For deterministic
                sigs, no salt is used: this means that for the same payload, the
                same sig is obtained (the advantage is that the sig is shorter).
            separator (optional): Character to separate the signature and the
                payload. It must not belong to the encoder alphabet and be ASCII
                (defaults to ".").
            encoder (optional): Encoder class to use (defaults to a Base64 URL
                safe encoder).

        Raises:
            ConversionError: A bytes parameter is not bytes and can't be converted
                to bytes.
            InvalidOptionError: A parameter is out of bounds.
        """
        if max_age is not None:
            personalisation = self._force_bytes(personalisation) + b'Timestamp'

        self._max_age: typing.Union[None, int, float, timedelta] = max_age

        super().__init__(
            secret,
            personalisation=personalisation,
            digest_size=digest_size,
            hasher=hasher,
            deterministic=deterministic,
            separator=separator,
            encoder=encoder,
        )

    def _proper_sign(self, data: bytes) -> bytes:
        """Sign given data with a (timestamp) signer producing a signature stream.

        The signature stream (salt, signature and/or timestamp) is encoded using
        the chosen encoder.

        Args:
            data: data to sign.

        Returns:
            A signature stream of the salt, signature, and timestamp, if corresponds.
        """
        if self._max_age is None:
            return self._sign(data)

        return self._sign_with_timestamp(data)

    def _proper_unsign(self, parts: SignedDataParts) -> bytes:
        """Unsign signed data properly with the corresponding signer.

        Args:
            parts: signed data parts to unsign.

        Returns:
            Original data.

        Raises:
            SignatureError: Signed data structure is not valid.
            InvalidSignatureError: Signed data signature is invalid.
            ExpiredSignatureError: Signed data signature has expired.
            DecodeError: Timestamp can't be decoded.
        """
        if self._max_age is None:
            return self._unsign(parts)

        return self._unsign_with_timestamp(parts, max_age=self._max_age)


class Blake2SerializerSignerBase(Blake2DualSignerBase, ABC):
    """Base class for a serializer signer that implements `dumps` and `loads`."""

    @abstractmethod
    def _dumps(self, data: typing.Any, **kwargs: typing.Any) -> bytes:
        """Dump data serializing it.

        Implement this method with all the tasks necessary to serialize data, such
        as encoding, compression, etc.

        Args:
            data: Data to serialize.

        Keyword Args:
            **kwargs: Additional keyword only arguments for the method.

        Returns:
            Serialized data.
        """

    @abstractmethod
    def _loads(self, dumped_data: bytes, **kwargs: typing.Any) -> typing.Any:
        """Load serialized data to recover it.

        Implement this method with all the tasks necessary to unserialize data,
        such as decoding, decompression, etc.

        Args:
            dumped_data: Data to unserialize.

        Keyword Args
            **kwargs: Additional keyword only arguments for the method.

        Returns:
            Original data.
        """

    @staticmethod
    def _read(file: typing.IO) -> typing.AnyStr:
        """Read data from a file.

        Args:
            file: file to read data from.

        Returns:
            File contents.

        Raises:
            FileError: File can't be read.
        """
        try:
            return file.read()
        except OSError as exc:
            raise FileError('file can not be read') from exc

    def _write(self, file: typing.IO, data: str) -> None:
        """Write data to file.

        Args:
            file: file to write data to.
            data: data to write to the file.

        Notes:
            The file can be either in text or binary mode, therefore given data
            is properly converted before writing.

        Raises:
            FileError: File can't be written.
            ConversionError: Data can't be converted to bytes (can happen when
                file is in binary mode).
        """
        data_ = data if file_mode_is_text(file) else self._force_bytes(data)

        try:
            file.write(data_)
        except OSError as exc:
            raise FileError('file can not be written') from exc



"""Utils module tests."""

import io
from datetime import datetime
from datetime import timedelta

from .. import utils


def test_b64encode_strips_padding() -> None:
    """Test b64encode strips padding correctly."""
    encoded = utils.b64encode(b'abcd')
    assert encoded == b'YWJjZA'

    encoded = utils.b64encode(b'abc')
    assert encoded == b'YWJj'


def test_b64decode_stripped_padding_works() -> None:
    """Test b64decode works without padding."""
    decoded = utils.b64decode(b'YWJjZA')
    assert decoded == b'abcd'

    decoded = utils.b64decode(b'YWJj')
    assert decoded == b'abc'


def test_b64decode_works_with_padding() -> None:
    """Test b64decode accepts padded value correctly."""
    decoded = utils.b64decode(b'YWJj==')
    assert decoded == b'abc'


def test_b32encode_strips_padding() -> None:
    """Test b32encode strips padding correctly."""
    encoded = utils.b32encode(b'abcd')
    assert encoded == b'MFRGGZA'

    encoded = utils.b32encode(b'abc')
    assert encoded == b'MFRGG'


def test_b32decode_stripped_padding_works() -> None:
    """Test b32decode works without padding."""
    decoded = utils.b32decode(b'MFRGGZA')
    assert decoded == b'abcd'

    decoded = utils.b32decode(b'MFRGG')
    assert decoded == b'abc'


def test_b32decode_works_with_padding() -> None:
    """Test b32decode accepts padded value correctly."""
    decoded = utils.b32decode(b'MFRGG===')
    assert decoded == b'abc'


def test_hexdecode_accepts_bytes() -> None:
    """Test hexdecode accepts bytes correctly."""
    decoded = utils.hexdecode(b'616263')
    assert decoded == b'abc'


def test_hexencode_accepts_bytes() -> None:
    """Test hexencode accepts bytes correctly."""
    encoded = utils.hexencode(b'abc')
    assert encoded == b'616263'


def test_timestamp_to_aware_datetime_accepts_int() -> None:
    """Test timestamp_to_aware_datetime accepts int timestamps."""
    timestamp = 1619064799
    converted_timestamp = utils.timestamp_to_aware_datetime(timestamp)

    assert isinstance(converted_timestamp, datetime)
    assert timedelta() == converted_timestamp.utcoffset()  # Aware in UTC
    # Technically, Python does some coercing, as in (1 == 1.0) is True, but just in case:
    assert timestamp == int(converted_timestamp.timestamp())
    assert float(timestamp) == converted_timestamp.timestamp()
    assert timestamp == converted_timestamp.timestamp()


def test_timestamp_to_aware_datetime_accepts_float() -> None:
    """Test timestamp_to_aware_datetime accepts float timestamps."""
    timestamp = 1619064799.965
    converted_timestamp = utils.timestamp_to_aware_datetime(timestamp)

    assert isinstance(converted_timestamp, datetime)
    assert timedelta() == converted_timestamp.utcoffset()  # Aware in UTC
    assert timestamp == converted_timestamp.timestamp()


def test_file_mode_is_text() -> None:
    """Test file_mode_is_text works correctly."""
    text_file = io.StringIO()
    assert utils.file_mode_is_text(text_file)

    bin_file = io.BytesIO()
    assert not utils.file_mode_is_text(bin_file)


def test_ordinal() -> None:
    """Test that the ordinal function works correctly."""
    assert '0th' == utils.ordinal(0)
    assert '1st' == utils.ordinal(1)
    assert '2nd' == utils.ordinal(2)
    assert '3rd' == utils.ordinal(3)
    assert '4th' == utils.ordinal(4)
    assert '12th' == utils.ordinal(12)
    assert '100th' == utils.ordinal(100)
    assert '101st' == utils.ordinal(101)
    assert '1003rd' == utils.ordinal(1003)
